"""Scrape MLB player data"""
from datetime import date, datetime, timedelta
from json.decoder import JSONDecodeError

from sqlalchemy.exc import DBAPIError, SQLAlchemyError

from vigorish.config.database import Player
from vigorish.util.request_url import request_url_with_retries
from vigorish.util.result import Result
from vigorish.util.string_helpers import fuzzy_match

MLB_PLAYER_SEARCH_URL = "http://lookup-service-prod.mlb.com/json/named.search_player_all.bam"
MLB_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S"


def scrape_mlb_player_info(db_session, name, bbref_id, game_date):
    return (
        get_search_url(name)
        .on_success(request_url_with_retries)
        .on_success(decode_json_response)
        .on_success(get_player_data, name, game_date)
        .on_success(parse_player_data, bbref_id, db_session)
    )


def get_search_url(name):
    split = name.split()
    if not split or len(split) <= 1:
        return Result.Fail(f"Name was not in an expected format: {name}")
    name_part = ("%20".join(split[1:])).upper()
    url = f"{MLB_PLAYER_SEARCH_URL}?sport_code='mlb'&name_part='{name_part}%25'&active_sw='Y'"
    return Result.Ok(url)


def decode_json_response(response):
    try:
        resp_json = response.json()
        query_results = resp_json["search_player_all"]["queryResults"]
        num_results = int(query_results["totalSize"])
        return Result.Ok((query_results, num_results))
    except (JSONDecodeError, KeyError) as e:
        error = f"Failed to decode HTTP response as JSON: {repr(e)}\n{e.response.text}"
        return Result.Fail(error)
    except ValueError:
        error = f"Failed to parse number of results from search response: {query_results}"
        return Result.Fail(error)


def get_player_data(results_tuple, name, game_date):
    results = results_tuple[0]
    num_results = results_tuple[1]
    if num_results > 1:
        player_list = results["row"]
        player_data = find_best_match(player_list, name, game_date)
    else:
        player_data = results["row"]
    return Result.Ok(player_data)


def find_best_match(player_list, name, game_date):
    player_id_name_map = {
        player["player_id"]: player["name_display_first_last"] for player in player_list
    }
    player_info_dict = {player["player_id"]: player for player in player_list}
    match_results = fuzzy_match(name, player_id_name_map)
    if len(match_results) == 1:
        return player_info_dict[match_results[0]["result"]]
    return compare_mlb_debut(match_results, player_info_dict, game_date)


def compare_mlb_debut(possible_matches, player_info_dict, game_date):
    info_dict_list = [
        player_info_dict[possible_match["result"]] for possible_match in possible_matches
    ]
    for player_info in info_dict_list:
        if not player_info["pro_debut_date"]:
            player_info["since_debut"] = timedelta.max.days
            continue
        player_mlb_debut = datetime.strptime(player_info["pro_debut_date"], MLB_DATE_FORMAT)
        player_info["since_debut"] = abs((game_date - player_mlb_debut.date()).days)
    info_dict_list.sort(key=lambda x: x["since_debut"])
    return info_dict_list[0]


def parse_player_data(player_info, bbref_id, db_session):
    try:
        feet = int(player_info["height_feet"])
        inches = int(player_info["height_inches"])
        height_total_inches = (feet * 12) + inches
    except ValueError:
        height_total_inches = 0

    try:
        debut = datetime.strptime(player_info["pro_debut_date"], MLB_DATE_FORMAT).date()
        birth_date = datetime.strptime(player_info["birth_date"], MLB_DATE_FORMAT).date()
    except ValueError:
        debut = date.min
        birth_date = date.min

    player_dict = {
        "name_first": player_info["name_first"],
        "name_last": player_info["name_last"],
        "name_given": player_info["name_first"],
        "bats": player_info["bats"],
        "throws": player_info["throws"],
        "weight": player_info["weight"],
        "height": height_total_inches,
        "debut": debut,
        "birth_year": birth_date.year,
        "birth_month": birth_date.month,
        "birth_day": birth_date.day,
        "birth_country": player_info["birth_country"],
        "birth_state": player_info["birth_state"],
        "birth_city": player_info["birth_city"],
        "bbref_id": bbref_id,
        "mlb_id": player_info["player_id"],
        "missing_mlb_id": False,
    }

    try:
        new_player = Player(**player_dict)
        db_session.add(new_player)
        db_session.commit()
        return Result.Ok(new_player)
    except (SQLAlchemyError, DBAPIError) as e:
        return Result.Fail(f"Error: {repr(e)}")



"""Reusable menu prompts to get various values/data types from the user."""
import subprocess

from bullet import Bullet, Check, colors, ScrollBar
from getch import pause

from vigorish.cli.components.data_set_check import DataSetCheck
from vigorish.cli.components.date_input import DateInput
from vigorish.cli.components.util import print_message
from vigorish.config.database import Season
from vigorish.constants import DATA_SET_NAMES_LONG, EMOJI_DICT, MENU_NUMBERS
from vigorish.enums import DataSet, VigFile
from vigorish.util.result import Result
from vigorish.util.string_helpers import wrap_text


def yes_no_prompt(prompt, wrap=True, max_line_len=70):
    if wrap:
        prompt = wrap_text(prompt, max_len=max_line_len)
    choices = {
        f"{MENU_NUMBERS.get(1)}  YES": True,
        f"{MENU_NUMBERS.get(2)}  NO": False,
    }
    prompt = Bullet(
        prompt,
        choices=[choice for choice in choices.keys()],
        bullet="",
        shift=1,
        indent=0,
        margin=2,
        bullet_color=colors.foreground["default"],
        word_color=colors.foreground["default"],
        word_on_switch=colors.bright(colors.foreground["cyan"]),
        background_color=colors.background["default"],
        background_on_switch=colors.background["default"],
    )
    print()
    choice_text = prompt.launch()
    return choices.get(choice_text)


def yes_no_cancel_prompt(prompt, wrap=True, max_line_len=70):
    if wrap:
        prompt = wrap_text(prompt, max_len=max_line_len)
    choices = {
        f"{MENU_NUMBERS.get(1)}  YES": True,
        f"{MENU_NUMBERS.get(2)}  NO": False,
        f"{MENU_NUMBERS.get(3)}  CANCEL": None,
    }
    prompt = Bullet(
        prompt,
        choices=[choice for choice in choices.keys()],
        bullet="",
        shift=1,
        indent=0,
        margin=2,
        bullet_color=colors.foreground["default"],
        word_color=colors.foreground["default"],
        word_on_switch=colors.bright(colors.foreground["cyan"]),
        background_color=colors.background["default"],
        background_on_switch=colors.background["default"],
    )
    choice_text = prompt.launch()
    choice_value = choices.get(choice_text)
    return Result.Fail("") if "CANCEL" in choice_text else Result.Ok(choice_value)


def single_date_prompt(prompt):
    user_date = None
    while not user_date:
        subprocess.run(["clear"])
        date_prompt = DateInput(prompt=prompt)
        result = date_prompt.launch()
        if result:
            user_date = result
    return user_date


def user_options_prompt(choices, prompt, wrap=True, clear_screen=True, auto_scroll=True):
    if wrap:
        prompt = wrap_text(prompt, 70)
    if auto_scroll and len(choices) > 8:
        scroll_choices = [choice for choice in choices.keys()]
        scroll_choices.insert(0, f"{EMOJI_DICT.get('BACK')} Return to Previous Menu")
        options_menu = ScrollBar(
            prompt,
            choices=scroll_choices,
            height=8,
            pointer="",
            shift=1,
            indent=0,
            margin=2,
            pointer_color=colors.foreground["default"],
            word_color=colors.foreground["default"],
            word_on_switch=colors.bright(colors.foreground["cyan"]),
            background_color=colors.background["default"],
            background_on_switch=colors.background["default"],
        )
    else:
        bullet_choices = [choice for choice in choices.keys()]
        if not auto_scroll:
            bullet_choices.insert(0, f"{EMOJI_DICT.get('BACK')} Return to Previous Menu")
        options_menu = Bullet(
            prompt,
            choices=bullet_choices,
            bullet="",
            shift=1,
            indent=0,
            margin=2,
            bullet_color=colors.foreground["default"],
            background_color=colors.foreground["default"],
            background_on_switch=colors.foreground["default"],
            word_color=colors.foreground["default"],
            word_on_switch=colors.bright(colors.foreground["cyan"]),
        )
    if clear_screen:
        subprocess.run(["clear"])
        print()
    choice_text = options_menu.launch()
    choice_value = choices.get(choice_text)
    return Result.Ok(choice_value) if choice_value else Result.Fail("")


def season_prompt(db_session, prompt=None):
    if not prompt:
        prompt = "Please select a MLB Season from the list below:"
    choices = {
        f"{MENU_NUMBERS.get(num)}  {season.year}": season
        for num, season in enumerate(Season.all_regular_seasons(db_session), start=1)
    }
    choices[f"{EMOJI_DICT.get('BACK')} Return to Previous Menu"] = None
    return user_options_prompt(choices, prompt)


def audit_report_season_prompt(audit_report, prompt=None):
    if not prompt:
        prompt = "Select an MLB season from the list below:"
    choices = {
        f"{MENU_NUMBERS.get(num)}  {year}": year
        for num, year in enumerate(audit_report.keys(), start=1)
    }
    choices[f"{EMOJI_DICT.get('BACK')} Return to Previous Menu"] = None
    return user_options_prompt(choices, prompt)


def data_sets_prompt(prompt, valid_data_sets=0, checked_data_sets=None):
    data_set_name_map = {v: k for k, v in DATA_SET_NAMES_LONG.items()}
    if not prompt:
        prompt = "Select one or multiple data sets from the list below:"
    if not valid_data_sets:
        valid_data_sets = int(DataSet.ALL)
    instructions = "(use SPACE BAR to select each data set, ENTER to confirm your selections)"
    choices = {
        f"{data_set_name_map[ds]}": ds
        for ds in DataSet
        if ds != DataSet.ALL and valid_data_sets & ds == ds
    }
    if checked_data_sets:
        checked_int = sum(int(ds) for ds in checked_data_sets.keys())
        checked_data_sets = [
            f"{data_set_name_map[ds]}"
            for ds in DataSet
            if ds != DataSet.ALL and checked_int & ds == ds
        ]
    ds_prompt = DataSetCheck(
        prompt=instructions,
        choices=[data_set for data_set in choices.keys()],
        checked_data_sets=checked_data_sets,
        check=EMOJI_DICT.get("CHECK", ""),
        shift=1,
        indent=0,
        margin=2,
        check_color=colors.foreground["default"],
        check_on_switch=colors.foreground["default"],
        background_color=colors.foreground["default"],
        background_on_switch=colors.foreground["default"],
        word_color=colors.foreground["default"],
        word_on_switch=colors.bright(colors.foreground["cyan"]),
    )
    data_sets = []
    while not data_sets:
        subprocess.run(["clear"])
        print_message(prompt, fg="bright_yellow", bold=True, underline=True)
        result = ds_prompt.launch()
        if not result:
            print_message("\nYou must select at least one file type!", fg="bright_red", bold=True)
            pause(message="Press any key to continue...")
            continue
        data_sets = {DATA_SET_NAMES_LONG[sel]: sel for sel in result}
    return data_sets


def file_types_prompt(prompt, valid_file_types=VigFile.ALL):
    if not prompt:
        prompt = "Select one or multiple file types from the list below:"
    choices = {
        f"{file_type}": file_type
        for file_type in VigFile
        if file_type != VigFile.ALL and int(file_type) & valid_file_types == file_type
    }
    instructions = "(use SPACE BAR to select each file type, ENTER to confirm your selections)"
    file_types_prompt = Check(
        prompt=instructions,
        choices=[file_type for file_type in choices.keys()],
        check=EMOJI_DICT.get("CHECK", ""),
        shift=1,
        indent=0,
        margin=2,
        check_color=colors.foreground["default"],
        check_on_switch=colors.foreground["default"],
        word_color=colors.foreground["default"],
        word_on_switch=colors.bright(colors.foreground["cyan"]),
        background_color=colors.background["default"],
        background_on_switch=colors.background["default"],
    )
    file_types = []
    while not file_types:
        subprocess.run(["clear"])
        print_message(prompt, fg="bright_yellow", bold=True, underline=True)
        result = file_types_prompt.launch()
        if not result:
            print_message("\nYou must select at least one file type!", fg="bright_red", bold=True)
            pause(message="Press any key to continue...")
            continue
        file_types = [choices[sel] for sel in result]
    return file_types


def select_game_prompt(game_ids, prompt=None, use_numbers=True, clear_screen=True):
    if not prompt:
        prompt = "Select a game from the list below:"
    choices = {
        f"{_get_menu_item_emoji(use_numbers, num)}  {game_id}": game_id
        for num, game_id in enumerate(game_ids, start=1)
    }
    choices[f"{EMOJI_DICT.get('BACK')} Return to Main Menu"] = None
    return user_options_prompt(choices, prompt, clear_screen=clear_screen)


def _get_menu_item_emoji(use_numbers, num=None):
    if use_numbers and num:
        return MENU_NUMBERS.get(num, str(num))
    return EMOJI_DICT.get("BLUE_DIAMOND")



from dataclasses import dataclass
from datetime import datetime, timezone

from dataclass_csv import accept_whitespaces, dateformat
from sqlalchemy import Column, DateTime, Float, ForeignKey, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

from vigorish.config.database import Base
from vigorish.util.dataclass_helpers import dict_from_dataclass, sanitize_row_dict
from vigorish.util.datetime_util import make_tzaware, TIME_ZONE_NEW_YORK
from vigorish.util.dt_format_strings import CSV_UTC, DT_AWARE
from vigorish.util.string_helpers import csv_sanitize


class PitchFx(Base):

    __tablename__ = "pitchfx"
    id = Column(Integer, primary_key=True)
    bb_game_id = Column(String)
    bbref_game_id = Column(String)
    pitch_app_id = Column(String)
    inning_id = Column(String)
    at_bat_id = Column(String)
    pitcher_id_mlb = Column(Integer)
    batter_id_mlb = Column(Integer)
    pitcher_id_bbref = Column(String)
    batter_id_bbref = Column(String)
    pitcher_team_id_bb = Column(String)
    opponent_team_id_bb = Column(String)
    p_throws = Column(String)
    stand = Column(String)
    pitch_id = Column(Integer)
    inning = Column(Integer)
    ab_total = Column(Integer)
    ab_count = Column(Integer)
    ab_id = Column(Integer)
    des = Column(String)
    strikes = Column(Integer)
    balls = Column(Integer)
    basic_type = Column(String)
    pdes = Column(String)
    mlbam_pitch_name = Column(String)
    start_speed = Column(Float)
    spin = Column(Float)
    zone_location = Column(Integer)
    sz_top = Column(Float)
    sz_bot = Column(Float)
    pfx_xdatafile = Column(Float)
    pfx_zdatafile = Column(Float)
    pfx_x = Column(Float)
    pfx_z = Column(Float)
    uncorrected_pfx_x = Column(Float)
    uncorrected_pfx_z = Column(Float)
    px = Column(Float)
    pz = Column(Float)
    pxold = Column(Float)
    pzold = Column(Float)
    park_sv_id = Column(String)
    game_start_time_utc = Column(DateTime)
    time_pitch_thrown_utc = Column(DateTime)
    seconds_since_game_start = Column(Integer)
    has_zone_location = Column(Integer)
    table_row_number = Column(Integer)
    is_patched = Column(Integer)
    is_duplicate_guid = Column(Integer)
    is_duplicate_pitch_number = Column(Integer)
    is_invalid_ibb = Column(Integer)
    is_out_of_sequence = Column(Integer)

    pitcher_id = Column(Integer, ForeignKey("player.id"), index=True)
    batter_id = Column(Integer, ForeignKey("player.id"), index=True)
    team_pitching_id = Column(Integer, ForeignKey("team.id"))
    team_batting_id = Column(Integer, ForeignKey("team.id"))
    game_status_id = Column(Integer, ForeignKey("scrape_status_game.id"))
    pitch_app_db_id = Column(Integer, ForeignKey("scrape_status_pitch_app.id"), index=True)
    date_id = Column(Integer, ForeignKey("scrape_status_date.id"))
    season_id = Column(Integer, ForeignKey("season.id"))

    pitcher = relationship("Player", foreign_keys=[pitcher_id])
    batter = relationship("Player", foreign_keys=[batter_id])
    team_pitching = relationship("Team", foreign_keys=[team_pitching_id])
    team_batting = relationship("Team", foreign_keys=[team_batting_id])
    game = relationship("GameScrapeStatus", foreign_keys=[game_status_id])
    pitch_app = relationship("PitchAppScrapeStatus", foreign_keys=[pitch_app_db_id])
    date = relationship("DateScrapeStatus", foreign_keys=[date_id])
    season = relationship("Season", foreign_keys=[season_id])

    @hybrid_property
    def game_date(self):
        return self.game_start_time.date()

    @hybrid_property
    def game_start_time(self):
        game_start_utc = make_tzaware(self.game_start_time_utc, use_tz=timezone.utc, localize=False)
        return make_tzaware(game_start_utc, use_tz=TIME_ZONE_NEW_YORK, localize=True)

    @hybrid_property
    def time_pitch_thrown(self):
        thrown_utc = make_tzaware(self.time_pitch_thrown_utc, use_tz=timezone.utc, localize=False)
        return make_tzaware(thrown_utc, use_tz=TIME_ZONE_NEW_YORK, localize=True)

    def as_csv_dict(self):
        self.pdes = csv_sanitize(self.pdes)
        return dict_from_dataclass(self, PitchFxCsvRow, date_format=CSV_UTC)

    @classmethod
    def get_csv_col_names(cls):
        return [name for name in PitchFxCsvRow.__dataclass_fields__.keys()]

    @classmethod
    def export_table_as_csv(cls, db_session):
        col_names = ",".join(cls.get_csv_col_names())
        csv_dicts = (obj.as_csv_dict() for obj in db_session.query(cls).all())
        csv_rows = (",".join(sanitize_row_dict(d, date_format=CSV_UTC)) for d in csv_dicts)
        yield col_names
        for row in csv_rows:
            yield row

    @classmethod
    def from_dict(cls, pfx_dict):
        game_start_str = pfx_dict.pop("game_start_time_str")
        pitch_thrown_str = pfx_dict.pop("time_pitch_thrown_str")
        game_start_time = datetime.strptime(game_start_str, DT_AWARE).astimezone(timezone.utc)
        time_pitch_thrown = datetime.strptime(pitch_thrown_str, DT_AWARE).astimezone(timezone.utc)
        seconds_since_game_start = int((time_pitch_thrown - game_start_time).total_seconds())
        pfx_dict["game_start_time_utc"] = game_start_time
        pfx_dict["time_pitch_thrown_utc"] = time_pitch_thrown
        pfx_dict["seconds_since_game_start"] = seconds_since_game_start
        pfx_dict["basic_type"] = pfx_dict.pop("type")
        pfx_dict["pitch_id"] = pfx_dict.pop("id")
        pfx_dict["pitcher_id_mlb"] = int(pfx_dict.pop("pitcher_id"))
        pfx_dict["batter_id_mlb"] = int(pfx_dict.pop("batter_id"))
        pfx_dict["zone_location"] = int(pfx_dict["zone_location"])
        pfx_dict["spin"] = round(pfx_dict["spin"], 1)
        pfx_dict["pfx_x"] = round(pfx_dict["pfx_x"], 2)
        pfx_dict["pfx_z"] = round(pfx_dict["pfx_z"], 2)
        pfx_dict["px"] = round(pfx_dict["px"], 2)
        pfx_dict["pz"] = round(pfx_dict["pz"], 2)
        pfx_dict.pop("play_guid", None)
        pfx_dict.pop("pitcher_name", None)
        pfx_dict.pop("pitch_con", None)
        pfx_dict.pop("norm_ht", None)
        pfx_dict.pop("tstart", None)
        pfx_dict.pop("vystart", None)
        pfx_dict.pop("ftime", None)
        pfx_dict.pop("x0", None)
        pfx_dict.pop("y0", None)
        pfx_dict.pop("z0", None)
        pfx_dict.pop("vx0", None)
        pfx_dict.pop("vy0", None)
        pfx_dict.pop("vz0", None)
        pfx_dict.pop("ax", None)
        pfx_dict.pop("ay", None)
        pfx_dict.pop("az", None)
        pfx_dict.pop("tm_spin", None)
        pfx_dict.pop("sb", None)
        return cls(**pfx_dict)


@accept_whitespaces
@dateformat(CSV_UTC)
@dataclass
class PitchFxCsvRow:
    id: int = None
    bb_game_id: str = None
    bbref_game_id: str = None
    pitch_app_id: str = None
    inning_id: str = None
    at_bat_id: str = None
    pitcher_id_mlb: int = None
    batter_id_mlb: int = None
    pitcher_id_bbref: str = None
    batter_id_bbref: str = None
    pitcher_team_id_bb: str = None
    opponent_team_id_bb: str = None
    p_throws: str = None
    stand: str = None
    pitch_id: int = None
    inning: int = None
    ab_total: int = None
    ab_count: int = None
    ab_id: int = None
    des: str = None
    strikes: int = 0
    balls: int = 0
    basic_type: str = None
    pdes: str = None
    mlbam_pitch_name: str = None
    start_speed: float = None
    spin: float = None
    zone_location: int = None
    sz_top: float = None
    sz_bot: float = None
    pfx_xdatafile: float = None
    pfx_zdatafile: float = None
    pfx_x: float = None
    pfx_z: float = None
    uncorrected_pfx_x: float = None
    uncorrected_pfx_z: float = None
    px: float = None
    pz: float = None
    pxold: float = None
    pzold: float = None
    park_sv_id: str = None
    game_start_time_utc: datetime = None
    time_pitch_thrown_utc: datetime = None
    seconds_since_game_start: int = 0
    has_zone_location: int = 0
    table_row_number: int = 0
    is_patched: int = 0
    is_duplicate_guid: int = 0
    is_duplicate_pitch_number: int = 0
    is_invalid_ibb: int = 0
    is_out_of_sequence: int = 0



class BBRefStartingLineupSlot:
    """Batting order and defensive position for a single player in a starting lineup."""

    player_id_br = ""
    bat_order = ""
    def_position = ""

    def as_dict(self):
        """Convert lineup details to a dictionary."""
        dict = {
            "player_id_br": "{}".format(self.player_id_br),
            "bat_order": int(self.bat_order),
            "def_position": "{}".format(self.def_position),
        }
        return dict



from datetime import datetime
from pathlib import Path

from dateutil import parser

from vigorish.config.database import DateScrapeStatus, PitchAppScrapeStatus, Season
from vigorish.config.database import Season_Game_PitchApp_View as Season_View
from vigorish.data.file_helper import FileHelper
from vigorish.data.html_storage import HtmlStorage
from vigorish.data.json_storage import JsonStorage
from vigorish.enums import DataSet, VigFile
from vigorish.util.regex import URL_ID_CONVERT_REGEX, URL_ID_REGEX
from vigorish.util.result import Result


class ScrapedData:
    def __init__(self, db_engine, db_session, config):
        self.db_engine = db_engine
        self.db_session = db_session
        self.config = config
        self.file_helper = FileHelper(config)
        self.html_storage = HtmlStorage(config, self.file_helper)
        self.json_storage = JsonStorage(config, self.file_helper)

    def get_local_folderpath(self, file_type, data_set, year):
        return self.file_helper.local_folderpath_dict[file_type][data_set].resolve(year=year)

    def get_s3_folderpath(self, file_type, data_set, year):
        return self.file_helper.s3_folderpath_dict[file_type][data_set].resolve(year=year)

    def check_s3_bucket(self):
        return self.file_helper.check_s3_bucket()

    def create_all_folderpaths(self, year):
        return self.file_helper.create_all_folderpaths(year)

    def html_stored_s3(self, data_set):
        return self.html_storage.html_stored_s3(data_set)

    def save_html(self, data_set, url_id, html):
        return self.html_storage.save_html(data_set, url_id, html)

    def get_html(self, data_set, url_id):
        return self.html_storage.get_html(data_set, url_id)

    def delete_html(self, data_set, url_id):
        return self.html_storage.delete_html(data_set, url_id)

    def save_json(self, data_set, parsed_data):
        return self.json_storage.save_json(data_set, parsed_data)

    def get_scraped_data(self, data_set, url_id):
        return self.json_storage.get_scraped_data(data_set, url_id)

    def save_patch_list(self, data_set, patch_list):
        return self.json_storage.save_patch_list(data_set, patch_list)

    def get_patch_list(self, data_set, url_id):
        return self.json_storage.get_patch_list(data_set, url_id)

    def apply_patch_list(self, data_set, url_id, scraped_data, boxscore=None):
        patch_list = self.get_patch_list(data_set, url_id)
        if not patch_list:
            return Result.Ok(scraped_data)
        return patch_list.apply(scraped_data, self.db_session, boxscore)

    def get_brooks_games_for_date(self, game_date):
        return self.json_storage.get_scraped_data(DataSet.BROOKS_GAMES_FOR_DATE, game_date)

    def get_brooks_pitch_logs_for_game(self, bb_game_id):
        return self.json_storage.get_scraped_data(DataSet.BROOKS_PITCH_LOGS, bb_game_id)

    def get_brooks_pitchfx_log(self, pitch_app_id):
        return self.json_storage.get_scraped_data(DataSet.BROOKS_PITCHFX, pitch_app_id)

    def get_bbref_games_for_date(self, game_date):
        return self.json_storage.get_scraped_data(DataSet.BBREF_GAMES_FOR_DATE, game_date)

    def get_bbref_boxscore(self, bbref_game_id):
        return self.json_storage.get_scraped_data(DataSet.BBREF_BOXSCORES, bbref_game_id)

    def save_combined_game_data(self, combined_data):
        return self.json_storage.save_combined_game_data(combined_data)

    def get_combined_game_data(self, bbref_game_id):
        return self.json_storage.get_combined_game_data(bbref_game_id)

    def get_all_brooks_pitch_logs_for_date(self, game_date):
        brooks_game_ids = DateScrapeStatus.get_all_brooks_game_ids_for_date(
            self.db_session, game_date
        )
        pitch_logs = []
        for game_id in brooks_game_ids:
            pitch_log = self.get_brooks_pitch_logs_for_game(game_id)
            if not pitch_log:
                continue
            pitch_logs.append(pitch_log)
        return pitch_logs

    def get_all_pitchfx_logs_for_game(self, bbref_game_id):
        p_app_ids = PitchAppScrapeStatus.get_all_scraped_pitch_app_ids_for_game_with_pitchfx_data(
            self.db_session, bbref_game_id
        )
        pitchfx_logs = [self.get_brooks_pitchfx_log(pitch_app_id) for pitch_app_id in p_app_ids]
        if any(not pfx_log for pfx_log in pitchfx_logs):
            return Result.Fail(f"Failed to retrieve all pitchfx logs for game {bbref_game_id}")
        return Result.Ok(pitchfx_logs)

    def get_all_brooks_dates_scraped_html(self, year):
        html_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.SCRAPED_HTML, data_set=DataSet.BROOKS_GAMES_FOR_DATE, year=year
        )
        scraped_dates = [
            Path(obj.key).stem
            for obj in self.file_helper.get_all_object_keys_in_s3_bucket()
            if html_folder in obj.key
        ]
        return scraped_dates

    def get_all_brooks_dates_scraped(self, year):
        json_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.PARSED_JSON, data_set=DataSet.BROOKS_GAMES_FOR_DATE, year=year
        )
        html_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.SCRAPED_HTML, data_set=DataSet.BROOKS_GAMES_FOR_DATE, year=year
        )
        scraped_dates = [
            parser.parse(Path(obj.key).stem[-10:])
            for obj in self.file_helper.get_all_object_keys_in_s3_bucket()
            if json_folder in obj.key and html_folder not in obj.key
        ]
        return Result.Ok(scraped_dates)

    def get_all_brooks_pitch_logs_scraped_html(self, year):
        html_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.SCRAPED_HTML, data_set=DataSet.BROOKS_PITCH_LOGS, year=year
        )
        scraped_gameids = [
            Path(obj.key).stem
            for obj in self.file_helper.get_all_object_keys_in_s3_bucket()
            if html_folder in obj.key
        ]
        return scraped_gameids

    def get_all_scraped_brooks_game_ids(self, year):
        json_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.PARSED_JSON, data_set=DataSet.BROOKS_PITCH_LOGS, year=year
        )
        html_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.SCRAPED_HTML, data_set=DataSet.BROOKS_PITCH_LOGS, year=year
        )
        scraped_gameids = [
            Path(obj.key).stem
            for obj in self.file_helper.get_all_object_keys_in_s3_bucket()
            if json_folder in obj.key and html_folder not in obj.key
        ]
        return Result.Ok(scraped_gameids)

    def get_all_pitchfx_pitch_app_ids_scraped_html(self, year):
        html_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.SCRAPED_HTML, data_set=DataSet.BROOKS_PITCHFX, year=year
        )
        scraped_pitch_app_ids = [
            Path(obj.key).stem
            for obj in self.file_helper.get_all_object_keys_in_s3_bucket()
            if html_folder in obj.key
        ]
        return scraped_pitch_app_ids

    def get_all_scraped_pitchfx_pitch_app_ids(self, year):
        json_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.PARSED_JSON, data_set=DataSet.BROOKS_PITCHFX, year=year
        )
        html_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.SCRAPED_HTML, data_set=DataSet.BROOKS_PITCHFX, year=year
        )
        scraped_pitch_app_ids = [
            Path(obj.key).stem
            for obj in self.file_helper.get_all_object_keys_in_s3_bucket()
            if json_folder in obj.key and html_folder not in obj.key
        ]
        return scraped_pitch_app_ids

    def get_all_bbref_dates_scraped_html(self, year):
        html_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.SCRAPED_HTML, data_set=DataSet.BBREF_GAMES_FOR_DATE, year=year
        )
        scraped_dates = [
            Path(obj.key).stem
            for obj in self.file_helper.get_all_object_keys_in_s3_bucket()
            if html_folder in obj.key
        ]
        return scraped_dates

    def get_all_bbref_dates_scraped(self, year):
        json_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.PARSED_JSON, data_set=DataSet.BBREF_GAMES_FOR_DATE, year=year
        )
        html_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.SCRAPED_HTML, data_set=DataSet.BBREF_GAMES_FOR_DATE, year=year
        )
        scraped_dates = [
            parser.parse(Path(obj.key).stem[-10:])
            for obj in self.file_helper.get_all_object_keys_in_s3_bucket()
            if json_folder in obj.key and html_folder not in obj.key
        ]
        return Result.Ok(scraped_dates)

    def get_all_bbref_game_ids_scraped_html(self, year):
        html_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.SCRAPED_HTML, data_set=DataSet.BBREF_BOXSCORES, year=year
        )
        scraped_game_ids = [
            Path(obj.key).stem
            for obj in self.file_helper.get_all_object_keys_in_s3_bucket()
            if html_folder in obj.key
        ]
        return scraped_game_ids

    def get_all_scraped_bbref_game_ids(self, year):
        json_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.PARSED_JSON, data_set=DataSet.BBREF_BOXSCORES, year=year
        )
        html_folder = self.file_helper.get_s3_folderpath(
            file_type=VigFile.SCRAPED_HTML, data_set=DataSet.BBREF_BOXSCORES, year=year
        )
        scraped_game_ids = [
            Path(obj.key).stem
            for obj in self.file_helper.get_all_object_keys_in_s3_bucket()
            if json_folder in obj.key and html_folder not in obj.key
        ]
        return Result.Ok(scraped_game_ids)

    def get_audit_report(self):
        all_seasons = Season.all_regular_seasons(self.db_session)
        total_games = self.get_total_games_for_all_seasons(all_seasons)
        scraped = self.get_all_bbref_game_ids_eligible_for_audit(all_seasons)
        successful = self.get_all_bbref_game_ids_all_pitchfx_logs_are_valid(all_seasons)
        failed = self.get_all_bbref_game_ids_combined_data_fail(all_seasons)
        pfx_error = self.get_all_bbref_game_ids_pitchfx_error(all_seasons)
        invalid_pfx = self.get_all_bbref_game_ids_invalid_pitchfx(all_seasons)
        mlb_seasons = list(
            set(
                list(scraped.keys())
                + list(successful.keys())
                + list(failed.keys())
                + list(pfx_error.keys())
                + list(invalid_pfx.keys())
            )
        )
        return {
            year: {
                "total_games": total_games[year],
                "scraped": scraped[year],
                "successful": successful[year],
                "failed": failed[year],
                "pfx_error": pfx_error[year],
                "invalid_pfx": invalid_pfx[year],
            }
            for year in mlb_seasons
            if scraped[year]
            or successful[year]
            or failed[year]
            or pfx_error[year]
            or invalid_pfx[year]
        }

    def get_total_games_for_all_seasons(self, all_seasons):
        return {s.year: s.total_games for s in all_seasons}

    def get_all_bbref_game_ids_eligible_for_audit(self, all_seasons):
        return {
            s.year: Season_View.get_all_bbref_game_ids_eligible_for_audit(self.db_engine, s.year)
            for s in all_seasons
        }

    def get_all_bbref_game_ids_combined_data_fail(self, all_seasons):
        return {s.year: s.get_all_bbref_game_ids_combined_data_fail() for s in all_seasons}

    def get_all_bbref_game_ids_pitchfx_error(self, all_seasons):
        return {
            s.year: Season_View.get_all_bbref_game_ids_pitchfx_error(self.db_engine, s.year)
            for s in all_seasons
        }

    def get_all_bbref_game_ids_invalid_pitchfx(self, all_seasons):
        return {
            s.year: Season_View.get_all_bbref_game_ids_invalid_pitchfx(self.db_engine, s.year)
            for s in all_seasons
        }

    def get_all_bbref_game_ids_all_pitchfx_logs_are_valid(self, all_seasons):
        return {
            s.year: Season_View.get_all_bbref_game_ids_all_pitchfx_logs_are_valid(
                self.db_engine, s.year
            )
            for s in all_seasons
        }

    def get_scraped_ids_from_local_folder(self, file_type, data_set, year):
        folderpath = self.get_local_folderpath(file_type, data_set, year)
        url_ids = [file.stem for file in Path(folderpath).glob("*.*")]
        return self.validate_url_ids(file_type, data_set, url_ids)

    def get_scraped_ids_from_database(self, data_set, year):
        get_scraped_ids_from_db_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.get_all_brooks_dates_scraped,
            DataSet.BROOKS_PITCH_LOGS: self.get_all_scraped_brooks_game_ids,
            DataSet.BROOKS_PITCHFX: self.get_all_scraped_pitchfx_pitch_app_ids,
            DataSet.BBREF_GAMES_FOR_DATE: self.get_all_bbref_dates_scraped,
            DataSet.BBREF_BOXSCORES: self.get_all_scraped_bbref_game_ids,
        }
        return get_scraped_ids_from_db_dict[data_set](year)

    def get_scraped_ids_from_s3(self, file_type, data_set, year):
        scraped_html_s3_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.get_all_brooks_dates_scraped_html,
            DataSet.BROOKS_PITCH_LOGS: self.get_all_brooks_pitch_logs_scraped_html,
            DataSet.BROOKS_PITCHFX: self.get_all_pitchfx_pitch_app_ids_scraped_html,
            DataSet.BBREF_GAMES_FOR_DATE: self.get_all_bbref_dates_scraped_html,
            DataSet.BBREF_BOXSCORES: self.get_all_bbref_game_ids_scraped_html,
        }
        url_ids = scraped_html_s3_dict[data_set](year)
        return self.validate_url_ids(file_type, data_set, url_ids)

    def validate_url_ids(self, file_type, data_set, url_ids):
        url_ids = [uid for uid in url_ids if URL_ID_REGEX[file_type][data_set].search(uid)]
        url_ids = self.convert_url_ids(file_type, data_set, url_ids)
        return sorted(url_ids)

    def convert_url_ids(self, file_type, data_set, url_ids):
        if data_set not in URL_ID_CONVERT_REGEX[file_type]:
            return url_ids
        convert_regex = URL_ID_CONVERT_REGEX[file_type][data_set]
        converted_url_ids = []
        for url_id in url_ids:
            match = convert_regex.search(url_id)
            if not match:
                raise ValueError(f"URL identifier is invalid: {url_id} ({data_set})")
            captured = match.groupdict()
            try:
                year = int(captured["year"])
                month = int(captured["month"])
                day = int(captured["day"])
                game_date = datetime(year, month, day)
            except Exception as e:
                error = f'Failed to parse date from url_id "{url_id} ({data_set})":\n{repr(e)}'
                raise ValueError(error)
            converted_url_ids.append(game_date)
        return converted_url_ids



"""Functions that enable reading/writing the config file."""
from vigorish.config.types.batch_job_settings import BatchJobSettings
from vigorish.config.types.batch_scrape_delay import BatchScrapeDelay
from vigorish.config.types.folder_path_setting import LocalFolderPathSetting, S3FolderPathSetting
from vigorish.config.types.url_scrape_delay import UrlScrapeDelay
from vigorish.enums import (
    CombinedDataStorageOption,
    ConfigType,
    DataSet,
    HtmlStorageOption,
    JsonStorageOption,
    ScrapeCondition,
    StatusReport,
)
from vigorish.util.list_helpers import report_dict


class ConfigSetting:
    setting_name: str

    def __init__(self, setting_name, config_dict):
        self.setting_name = setting_name
        self.config_dict = config_dict

    @property
    def setting_name_title(self):
        return " ".join(self.setting_name.split("_")).title()

    @property
    def data_type(self):
        return ConfigType.NONE

    @property
    def description(self):
        return self.config_dict.get("DESCRIPTION", "")

    @property
    def possible_values(self):
        return None

    @property
    def is_same_for_all_data_sets(self):
        return (
            True
            if self.same_value_for_all_data_sets_is_required
            else self.config_dict.get("SAME_SETTING_FOR_ALL_DATA_SETS")
        )

    @is_same_for_all_data_sets.setter
    def is_same_for_all_data_sets(self, is_same):
        if not self.same_value_for_all_data_sets_is_required:
            self.config_dict["SAME_SETTING_FOR_ALL_DATA_SETS"] = is_same

    @property
    def cannot_be_disabled(self):
        return self.setting_name in ["URL_SCRAPE_DELAY"]

    @property
    def same_value_for_all_data_sets_is_required(self):
        return self.setting_name in [
            "STATUS_REPORT",
            "S3_BUCKET",
            "SCRAPED_DATA_COMBINE_CONDITION",
            "COMBINED_DATA_STORAGE",
            "COMBINED_DATA_LOCAL_FOLDER_PATH",
            "COMBINED_DATA_S3_FOLDER_PATH",
            "DB_BACKUP_FOLDER_PATH",
        ]

    @property
    def current_settings_report(self):
        if self.is_same_for_all_data_sets:
            settings_dict = {"ALL_DATA_SETS": self.current_setting(DataSet.ALL)}
        else:
            settings_dict = {
                data_set.name: self.current_setting(data_set)
                for data_set in DataSet
                if data_set != DataSet.ALL
            }
        return report_dict(settings_dict, title="", title_prefix="", title_suffix="")

    def current_setting(self, data_set: DataSet):
        return (
            self.config_dict.get(DataSet.ALL.name)
            if self.is_same_for_all_data_sets
            else self.config_dict.get(data_set.name)
        )


class StringConfigSetting(ConfigSetting):
    def __repr__(self):
        return (
            "<StringConfigSetting "
            f"setting={self.setting_name},"
            f"value={self.current_settings_report}>"
        )

    @property
    def data_type(self):
        return ConfigType.STRING


class PathConfigSetting(ConfigSetting):
    def __repr__(self):
        return (
            "<PathConfigSetting "
            f"setting={self.setting_name},"
            f"value={self.current_settings_report}>"
        )

    @property
    def data_type(self):
        return ConfigType.PATH

    def current_setting(self, data_set):
        current_setting = super().current_setting(data_set)
        return self.get_object(self.config_dict.get("CLASS_NAME"), current_setting, data_set)

    @staticmethod
    def get_object(class_name, current_setting, data_set):
        if class_name == "LocalFolderPathSetting":
            return LocalFolderPathSetting(path=current_setting, data_set=data_set)
        if class_name == "S3FolderPathSetting":
            return S3FolderPathSetting(path=current_setting, data_set=data_set)
        return None


class EnumConfigSetting(ConfigSetting):
    def __repr__(self):
        return (
            "<EnumConfigSetting "
            f"setting={self.setting_name},"
            f"value={self.current_settings_report}>"
        )

    @property
    def data_type(self):
        return ConfigType.ENUM

    @property
    def possible_values(self):
        enum_name = self.config_dict.get("ENUM_NAME")
        if enum_name == "ScrapeCondition":
            return [member for member in ScrapeCondition]
        if enum_name == "HtmlStorageOption":
            return [member for member in HtmlStorageOption]
        if enum_name == "JsonStorageOption":
            return [member for member in JsonStorageOption]
        if enum_name == "CombinedDataStorageOption":
            return [member for member in CombinedDataStorageOption]
        if enum_name == "StatusReport":
            return [member for member in StatusReport]
        return []

    def current_setting(self, data_set):
        current_setting = super().current_setting(data_set)
        return self.get_enum(self.config_dict.get("ENUM_NAME"), current_setting)

    @staticmethod
    def get_enum(enum_name, value):
        if enum_name == "ScrapeCondition":
            return ScrapeCondition[value]
        if enum_name == "HtmlStorageOption":
            return HtmlStorageOption[value]
        if enum_name == "JsonStorageOption":
            return JsonStorageOption[value]
        if enum_name == "CombinedDataStorageOption":
            return CombinedDataStorageOption[value]
        if enum_name == "StatusReport":
            return StatusReport[value]
        return None


class NumericConfigSetting(ConfigSetting):
    def __repr__(self):
        return (
            "<NumericConfigSetting "
            f"setting={self.setting_name},"
            f"value={self.current_settings_report}>"
        )

    @property
    def data_type(self):
        return ConfigType.NUMERIC

    def current_setting(self, data_set):
        current_setting = super().current_setting(data_set)
        return self.get_object(self.config_dict.get("CLASS_NAME"), current_setting)

    @staticmethod
    def get_object(class_name, config_dict):
        if class_name == "UrlScrapeDelay":
            return UrlScrapeDelay.from_config(config_dict)
        if class_name == "BatchJobSettings":
            return BatchJobSettings.from_config(config_dict)
        if class_name == "BatchScrapeDelay":
            return BatchScrapeDelay.from_config(config_dict)
        return None



from pathlib import Path

from vigorish.config.database import DateScrapeStatus, GameScrapeStatus, Season
from vigorish.enums import DataSet
from vigorish.util.dt_format_strings import DATE_ONLY, DATE_ONLY_TABLE_ID
from vigorish.util.result import Result
from vigorish.util.string_helpers import validate_bbref_game_id_list


def update_bbref_games_for_date_single_date(db_session, season, games_for_date):
    game_date = games_for_date.game_date
    unscraped_bbref_dates = DateScrapeStatus.get_all_bbref_unscraped_dates_for_season(
        db_session, season.id
    )
    update_bbref_dates = set([game_date]) & set(unscraped_bbref_dates)
    if not update_bbref_dates:
        return Result.Ok()
    result = update_status_bbref_games_for_date(db_session, games_for_date)
    if result.failure:
        return result
    new_bbref_game_ids = [Path(url).stem for url in games_for_date.boxscore_urls]
    result = create_game_status_records(db_session, season, new_bbref_game_ids)
    if result.failure:
        return result
    return Result.Ok()


def update_data_set_bbref_games_for_date(scraped_data, db_session, season):
    result = scraped_data.get_all_bbref_dates_scraped(season.year)
    if result.failure:
        return result
    scraped_bbref_dates = result.value
    unscraped_bbref_dates = DateScrapeStatus.get_all_bbref_unscraped_dates_for_season(
        db_session, season.id
    )
    update_bbref_dates = set(scraped_bbref_dates) & set(unscraped_bbref_dates)
    if not update_bbref_dates:
        return Result.Ok()
    result = update_bbref_games_for_date_list(scraped_data, db_session, update_bbref_dates)
    if result.failure:
        return result
    new_bbref_game_ids = result.value
    return create_game_status_records(db_session, season, new_bbref_game_ids)


def update_bbref_games_for_date_list(scraped_data, db_session, scraped_bbref_dates):
    season = None
    for game_date in scraped_bbref_dates:
        if not season:
            season = Season.find_by_year(db_session, game_date.year)
        games_for_date = scraped_data.get_bbref_games_for_date(game_date)
        if not games_for_date:
            date_str = game_date.strftime(DATE_ONLY_TABLE_ID)
            error = f"Failed to retrieve {DataSet.BBREF_GAMES_FOR_DATE} (URL ID: {date_str})"
            return Result.Fail(error)
        result = update_status_bbref_games_for_date(db_session, games_for_date)
        if result.failure:
            return result
        game_ids = [Path(url).stem for url in games_for_date.boxscore_urls]
        result = create_game_status_records(db_session, season, game_ids)
        if result.failure:
            return result
        db_session.commit()
    return Result.Ok()


def update_status_bbref_games_for_date(db_session, games_for_date):
    try:
        date_id = games_for_date.game_date.strftime(DATE_ONLY_TABLE_ID)
        date_status = db_session.query(DateScrapeStatus).get(date_id)
        if not date_status:
            date_str = games_for_date.game_date.strftime(DATE_ONLY)
            error = f"scrape_status_date does not contain an entry for date: {date_str}"
            return Result.Fail(error)
        date_status.scraped_daily_dash_bbref = 1
        date_status.game_count_bbref = games_for_date.game_count
        return Result.Ok()
    except Exception as e:
        return Result.Fail(f"Error: {repr(e)}")


def create_game_status_records(db_session, season, new_bbref_game_ids):
    game_status_bbref_ids = GameScrapeStatus.get_all_bbref_game_ids(db_session, season.id)
    missing_bbref_game_ids = set(new_bbref_game_ids).difference(set(game_status_bbref_ids))
    if not missing_bbref_game_ids:
        return Result.Ok()
    for game_dict in validate_bbref_game_id_list(missing_bbref_game_ids):
        try:
            game_date = game_dict["game_date"]
            date_status = DateScrapeStatus.find_by_date(db_session, game_date)
            game_status = GameScrapeStatus()
            game_status.game_date = game_date
            game_status.bbref_game_id = game_dict["game_id"]
            game_status.scrape_status_date_id = date_status.id
            game_status.season_id = season.id
            db_session.add(game_status)
        except Exception as e:
            return Result.Fail(f"Error: {repr(e)}")
    return Result.Ok()



from datetime import datetime

from events import Events

from vigorish.config.database import (
    GameScrapeStatus,
    PitchAppScrapeStatus,
    PitchFx,
    PlayerId,
    Season,
    Team,
)
from vigorish.data.all_game_data import AllGameData
from vigorish.tasks.base import Task
from vigorish.util.dt_format_strings import DATE_ONLY_TABLE_ID
from vigorish.util.result import Result
from vigorish.util.string_helpers import get_bbref_team_id, validate_bbref_game_id


class AddPitchFxToDatabase(Task):
    def __init__(self, app):
        super().__init__(app)
        self._season_id_map = {}
        self._team_id_map = {}
        self._player_id_map = {}
        self._game_id_map = {}
        self._pitch_app_id_map = {}
        self.events = Events(
            (
                "add_pitchfx_to_db_start",
                "add_pitchfx_to_db_progress",
                "add_pitchfx_to_db_complete",
            )
        )

    @property
    def season_id_map(self):
        if self._season_id_map:
            return self._season_id_map
        self._season_id_map = Season.regular_season_map(self.db_session)
        return self._season_id_map

    def get_team_id_map(self, year):
        if year in self._team_id_map:
            return self._team_id_map[year]
        self._team_id_map[year] = Team.get_team_id_map_for_year(self.db_session, year)
        return self._team_id_map[year]

    @property
    def player_id_map(self):
        if self._player_id_map:
            return self._player_id_map
        self._player_id_map = PlayerId.get_player_id_map(self.db_session)
        return self._player_id_map

    @property
    def game_id_map(self):
        if self._game_id_map:
            return self._game_id_map
        self._game_id_map = GameScrapeStatus.get_game_id_map(self.db_session)
        return self._game_id_map

    @property
    def pitch_app_id_map(self):
        if self._pitch_app_id_map:
            return self._pitch_app_id_map
        self._pitch_app_id_map = PitchAppScrapeStatus.get_pitch_app_id_map(self.db_session)
        return self._pitch_app_id_map

    def execute(self, audit_report, year=None):
        self.audit_report = audit_report
        return self.add_pfx_for_year(year) if year else self.add_all_pfx()

    def add_all_pfx(self):
        valid_years = [year for year, results in self.audit_report.items() if results["successful"]]
        for year in valid_years:
            self.add_pfx_for_year(year)
        return Result.Ok()

    def add_pfx_for_year(self, year):
        report_for_season = self.audit_report.get(year)
        if not report_for_season:
            return Result.Fail(f"Audit report could not be generated for MLB Season {year}")
        game_ids = report_for_season.get("successful")
        if not game_ids:
            error = f"No games for MLB Season {year} qualify to have PitchFx data imported."
            return Result.Fail(error)
        self.events.add_pitchfx_to_db_start(year, game_ids)
        for num, game_id in enumerate(game_ids, start=1):
            self.add_pitchfx_to_database(game_id)
            self.events.add_pitchfx_to_db_progress(num, year, game_id)
        self.events.add_pitchfx_to_db_complete(year)
        return Result.Ok()

    def add_pitchfx_to_database(self, game_id):
        all_game_data = AllGameData(self.app, game_id)
        for pitch_app_id, pfx_dict_list in all_game_data.get_all_pitchfx().items():
            pitch_app = PitchAppScrapeStatus.find_by_pitch_app_id(self.db_session, pitch_app_id)
            if not pitch_app:
                error = f"PitchFx import aborted! Pitch app '{pitch_app_id}' not found in database"
                return Result.Fail(error)
            if pitch_app.imported_pitchfx:
                continue
            for pfx_dict in pfx_dict_list:
                pfx = PitchFx.from_dict(pfx_dict)
                pfx = self.update_pitchfx_relationships(pfx)
                self.db_session.add(pfx)
            pitch_app.imported_pitchfx = 1
        self.db_session.commit()

    def update_pitchfx_relationships(self, pfx):
        game_date = self.get_game_date_from_bbref_game_id(pfx.bbref_game_id)
        pitcher_team_id_br = get_bbref_team_id(pfx.pitcher_team_id_bb)
        opponent_team_id_br = get_bbref_team_id(pfx.opponent_team_id_bb)
        pfx.pitcher_id = self.player_id_map[pfx.pitcher_id_mlb]
        pfx.batter_id = self.player_id_map[pfx.batter_id_mlb]
        pfx.team_pitching_id = self.get_team_id_map(game_date.year)[pitcher_team_id_br]
        pfx.team_batting_id = self.get_team_id_map(game_date.year)[opponent_team_id_br]
        pfx.season_id = self.season_id_map[game_date.year]
        pfx.date_id = self.get_date_status_id_from_game_date(game_date)
        pfx.game_status_id = self.game_id_map[pfx.bbref_game_id]
        pfx.pitch_app_db_id = self.pitch_app_id_map[pfx.pitch_app_id]
        return pfx

    def get_game_date_from_bbref_game_id(self, bbref_game_id):
        game_date = validate_bbref_game_id(bbref_game_id).value["game_date"]
        return datetime(game_date.year, game_date.month, game_date.day)

    def get_date_status_id_from_game_date(self, game_date):
        return game_date.strftime(DATE_ONLY_TABLE_ID)



"""Menu item that allows the user to initialize/reset the database."""
import subprocess
import time
from datetime import datetime
from pathlib import Path

from getch import pause

from vigorish.cli.components.prompts import yes_no_prompt
from vigorish.cli.components.util import print_heading, print_message, shutdown_cli_immediately
from vigorish.cli.menu_item import MenuItem
from vigorish.cli.menu_items.admin_tasks.update_player_id_map import UpdatePlayerIdMap
from vigorish.config.database import (
    db_setup_complete,
    initialize_database,
    reset_database_connection,
    Season,
)
from vigorish.constants import EMOJI_DICT
from vigorish.enums import DataSet
from vigorish.tasks.import_scraped_data import ImportScrapedDataTask
from vigorish.util.result import Result

SETUP_HEADING = (
    "Before you can begin scraping data, you must initialize the database with initial player, "
    "team and season data."
)
SETUP_MESSAGE = "Select YES to initialize the database\nSelect NO to return to the previous menu"
RESET_MESSAGE = "Would you like to reset the database with initial player, team and season data?"
WARNING = (
    "WARNING! All existing data will be deleted if you choose to reset the database. This "
    "action cannot be undone."
)
IMPORT_SCRAPED_DATA_MESSAGE = (
    "Would you like to update the database using files from previous vigorish installations? "
    "Files from any MLB season and any data set found in the location specified by the "
    "JSON_LOCAL_FOLDER_PATH config setting will be used for this step.\n"
)
IMPORT_SCRAPED_DATA_PROMPT = (
    "Select YES to import all data in the location above\n"
    "Select NO to return to the previous menu"
)
DB_INITIALIZED = "Database has been successfully initialized.\n"


class SetupDatabase(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.update_id_map_task = UpdatePlayerIdMap(self.app)
        self.import_data_task = ImportScrapedDataTask(self.app)
        self.db_initialized = db_setup_complete(self.db_engine, self.db_session)
        self.menu_item_text = "Reset Database" if self.db_initialized else "Setup Database"
        self.menu_item_emoji = EMOJI_DICT["BOMB"] if self.db_initialized else EMOJI_DICT["DIZZY"]

    def launch(self):
        restart_required = self.db_initialized
        if not self.prompt_user_run_task():
            return Result.Ok(self.exit_menu)
        return (
            self.update_player_id_map()
            .on_success(self.update_database_connection)
            .on_success(self.create_and_populate_database_tables)
            .on_success(self.import_scraped_data, restart_required)
        )

    def prompt_user_run_task(self):
        if self.db_initialized:
            message = WARNING
            message_color = "bright_red"
            prompt = RESET_MESSAGE
        else:
            message = SETUP_HEADING
            message_color = None
            prompt = SETUP_MESSAGE
        subprocess.run(["clear"])
        print_heading(self.get_heading("Run Task?"), fg="bright_yellow")
        print_message(message, fg=message_color)
        return yes_no_prompt(prompt, wrap=False)

    def get_heading(self, current_action):
        return (
            f"Reset Database: {current_action}"
            if self.db_initialized
            else f"Setup Database: {current_action}"
        )

    def update_player_id_map(self):
        if not self.db_initialized:
            return Result.Ok()
        result = Season.is_date_in_season(self.db_session, datetime.now())
        if result.failure:
            return Result.Ok()
        subprocess.run(["clear"])
        return self.update_id_map_task.launch()

    def update_database_connection(self):
        if not self.db_initialized:
            return Result.Ok()
        result = reset_database_connection(self.app)
        if result.failure:
            return result
        self.app = result.value
        return Result.Ok()

    def create_and_populate_database_tables(self):
        subprocess.run(["clear"])
        print_heading(self.get_heading("In Progress"), fg="bright_yellow")
        result = initialize_database(self.app)
        if result.success:
            subprocess.run(["clear"])
            print_heading(self.get_heading("Complete"), fg="bright_yellow")
            print_message(DB_INITIALIZED, fg="bright_green", bold=True)
            pause(message="Press any key to continue...")
        return result

    def import_scraped_data(self, restart_required):
        if not self.import_scraped_data_prompt():
            return self.setup_complete(restart_required)
        result = self.import_data_task.execute(overwrite_existing=True)
        if result.error:
            print_message(result.error, fg="bright_red")
            pause(message="Press any key to continue...")
        else:
            time.sleep(2)
        return self.setup_complete(restart_required)

    def import_scraped_data_prompt(self):
        local_folder_path = self.config.all_settings.get("JSON_LOCAL_FOLDER_PATH")
        example_folder = local_folder_path.current_setting(data_set=DataSet.BBREF_BOXSCORES)
        root_folder = Path(example_folder.resolve(year=2019)).parent.parent
        current_setting = f"JSON_LOCAL_FOLDER_PATH: {root_folder}"
        subprocess.run(["clear"])
        print_heading(self.get_heading("Import Local JSON Folder?"), fg="bright_yellow")
        print_message(IMPORT_SCRAPED_DATA_MESSAGE)
        print_message(current_setting, wrap=False)
        return yes_no_prompt(IMPORT_SCRAPED_DATA_PROMPT, wrap=False)

    def setup_complete(self, restart_required):
        if restart_required:
            shutdown_cli_immediately()
        return Result.Ok(self.exit_menu)



class BBRefTeamLinescoreTotals:
    """Total runs, hits and errors credited to a team in one game."""

    team_id_br = ""
    total_runs = ""
    total_hits = ""
    total_errors = ""

    def as_dict(self):
        """Convert total runs, hits and errors for one team in a single game to a dictionary."""
        dict = {
            "team_id_br": "{}".format(self.team_id_br),
            "total_runs": int(self.total_runs),
            "total_hits": int(self.total_hits),
            "total_errors": int(self.total_errors),
        }
        return dict



from dataclasses import dataclass
from typing import Any

from vigorish.util.list_helpers import as_dict_list


@dataclass
class BBRefBatStats:
    """Statistics for all plate appearances that occurred for one player in a single game."""

    details: Any
    player_id_br: str = "0"
    player_team_id_br: str = "0"
    opponent_team_id_br: str = "0"
    at_bats: str = "0"
    runs_scored: str = "0"
    hits: str = "0"
    rbis: str = "0"
    bases_on_balls: str = "0"
    strikeouts: str = "0"
    plate_appearances: str = "0"
    avg_to_date: str = "0"
    obp_to_date: str = "0"
    slg_to_date: str = "0"
    ops_to_date: str = "0"
    total_pitches: str = "0"
    total_strikes: str = "0"
    wpa_bat: str = "0"
    avg_lvg_index: str = "0"
    wpa_bat_pos: str = "0"
    wpa_bat_neg: str = "0"
    re24_bat: str = "0"

    def as_dict(self):
        """Convert batting statistics for one player in a single game to a dictionary."""
        return dict(
            player_id_br=self.player_id_br,
            player_team_id_br=self.player_team_id_br,
            opponent_team_id_br=self.opponent_team_id_br,
            at_bats=int(self.at_bats),
            runs_scored=int(self.runs_scored),
            hits=int(self.hits),
            rbis=int(self.rbis),
            bases_on_balls=int(self.bases_on_balls),
            strikeouts=int(self.strikeouts),
            plate_appearances=int(self.plate_appearances),
            avg_to_date=float(self.avg_to_date),
            obp_to_date=float(self.obp_to_date),
            slg_to_date=float(self.slg_to_date),
            ops_to_date=float(self.ops_to_date),
            total_pitches=int(self.total_pitches),
            total_strikes=int(self.total_strikes),
            wpa_bat=float(self.wpa_bat),
            avg_lvg_index=float(self.avg_lvg_index),
            wpa_bat_pos=float(self.wpa_bat_pos),
            wpa_bat_neg=float(self.wpa_bat_neg),
            re24_bat=float(self.re24_bat),
            details=as_dict_list(self.details),
        )



def calc_pitch_mix(pitch_count_dict, total_pitches):
    pitch_types = []
    mix_total = 0
    for col_name, val in pitch_count_dict.items():
        if not ("_count" in col_name):
            continue
        if val == 0:
            continue
        pitch_type = col_name.split("_")[0].upper()
        pitch_percent = val / float(total_pitches)
        if pitch_percent < 0.01:
            continue
        mix_total += val
        pitch_types.append({"type": pitch_type, "percent": pitch_percent, "count": val})
    pitch_types.append({"type": "ALL", "percent": 1, "count": mix_total})
    pitch_types.sort(key=lambda x: x["percent"], reverse=True)
    return {x["type"]: {"percent": x["percent"], "count": x["count"]} for x in pitch_types}



import shutil
import subprocess
from collections import defaultdict
from datetime import datetime

from getch import pause
from halo import Halo

from vigorish.cli.components import print_message
from vigorish.config.database import ScrapeError
from vigorish.constants import FAKE_SPINNER, JOB_SPINNER_COLORS
from vigorish.enums import DataSet, JobStatus, StatusReport
from vigorish.scrape.bbref_boxscores.scrape_task import ScrapeBBRefBoxscores
from vigorish.scrape.bbref_games_for_date.scrape_task import ScrapeBBRefGamesForDate
from vigorish.scrape.brooks_games_for_date.scrape_task import ScrapeBrooksGamesForDate
from vigorish.scrape.brooks_pitch_logs.scrape_task import ScrapeBrooksPitchLogs
from vigorish.scrape.brooks_pitchfx.scrape_task import ScrapeBrooksPitchFx
from vigorish.status.report_status import report_date_range_status, report_season_status
from vigorish.util.dt_format_strings import DATE_ONLY_2
from vigorish.util.result import Result
from vigorish.util.sys_helpers import node_is_installed, node_modules_folder_exists

JOB_STATUS_TEXT_COLOR = {"scraped": "bright_green", "skipped": "blue"}
NODEJS_INSTALL_ERROR = (
    "Node.js installation cannot be located. Please install a recent, stable "
    "version and install dependencies from npm, see github repo for detailed "
    "intructions."
)
NPM_PACKAGES_INSTALL_ERROR = (
    "Nightmare is not installed, you must install it and other node dependencies in "
    "order to scrape any data."
)


class JobRunner:
    def __init__(self, db_job, db_session, config, scraped_data):
        self.db_job = db_job
        self.job_id = self.db_job.id
        self.job_name = self.db_job.name
        self.data_sets = self.db_job.data_sets
        self.start_date = self.db_job.start_date
        self.end_date = self.db_job.end_date
        self.season = self.db_job.season
        self.db_session = db_session
        self.config = config
        self.scraped_data = scraped_data
        self.spinners = defaultdict(lambda: Halo(spinner=FAKE_SPINNER))
        self.task_results = []
        self.status_report = self.config.get_current_setting("STATUS_REPORT", DataSet.ALL)

    def execute(self):
        result = self.initialize()
        if result.failure:
            return self.job_failed(result)
        for task_number, data_set in enumerate(self.data_sets, start=1):
            self.report_task_results()
            self.update_spinner(data_set, task_number)
            result = self.get_scrape_task_for_data_set(data_set).execute()
            if result.failure:
                if "skip" in result.error:
                    self.log_result_data_set_skipped(data_set, task_number)
                    self.spinners[data_set].stop()
                    continue
                return self.job_failed(result, data_set)
            self.log_result_data_set_scraped(data_set, task_number)
            self.spinners[data_set].stop()
        self.report_task_results()
        result = self.show_status_report()
        pause(message="Press any key to continue...")
        if result.failure:
            return self.job_failed(result)
        return self.job_succeeded()

    def get_scrape_task_for_data_set(self, data_set):
        task_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: ScrapeBrooksGamesForDate,
            DataSet.BROOKS_PITCH_LOGS: ScrapeBrooksPitchLogs,
            DataSet.BROOKS_PITCHFX: ScrapeBrooksPitchFx,
            DataSet.BBREF_GAMES_FOR_DATE: ScrapeBBRefGamesForDate,
            DataSet.BBREF_BOXSCORES: ScrapeBBRefBoxscores,
        }
        return task_dict[data_set](self.db_job, self.db_session, self.config, self.scraped_data)

    def update_spinner(self, data_set, task_number):
        self.spinners[data_set].text = self.scrape_task_started_text(data_set, task_number)
        self.spinners[data_set].color = JOB_SPINNER_COLORS[data_set]
        self.spinners[data_set].start()
        self.spinners[data_set].stop_and_persist(self.spinners[data_set].frame(), "")

    def log_result_data_set_scraped(self, data_set, task_number):
        self.task_results.append(self.scrape_task_complete_text(data_set, task_number))

    def log_result_data_set_skipped(self, data_set, task_number):
        self.task_results.append(self.scrape_task_skipped_text(data_set, task_number))

    def scrape_task_complete_text(self, data_set, task_number):
        scraped = f"Scraped data set: {data_set.name} (Task {task_number}/{len(self.data_sets)})"
        return (scraped, JOB_STATUS_TEXT_COLOR["scraped"])

    def scrape_task_skipped_text(self, data_set, task_number):
        skipped = f"Skipped data set: {data_set.name} (Task {task_number}/{len(self.data_sets)})"
        return (skipped, JOB_STATUS_TEXT_COLOR["skipped"])

    def scrape_task_started_text(self, data_set, task_number):
        return f"Scraping data set: {data_set.name} (Task {task_number}/{len(self.data_sets)})"

    def report_task_results(self):
        subprocess.run(["clear"])
        start_date_str = self.start_date.strftime(DATE_ONLY_2)
        end_date_str = self.end_date.strftime(DATE_ONLY_2)
        job_name_id = f"Job Name: {self.job_name} (ID: {self.job_id.upper()})"
        if self.job_name == self.job_id:
            job_name_id = f"Job ID: {self.job_id.upper()}"
        job_heading = f"{job_name_id} Date Range: {start_date_str}-{end_date_str}\n"
        print_message(job_heading, wrap=False, fg="bright_yellow", bold=True, underline=True)
        if not self.task_results:
            return
        for task_result in self.task_results:
            print_message(task_result[0], fg=task_result[1])

    def initialize(self):
        errors = []
        if not node_is_installed():
            errors.append(NODEJS_INSTALL_ERROR)
        elif not node_modules_folder_exists():
            errors.append(NPM_PACKAGES_INSTALL_ERROR)
        result = self.check_url_delay_settings()
        if result.failure:
            errors.append(result.error)
        result = self.check_s3_bucket() if self.s3_bucket_required() else Result.Ok()
        if result.failure:
            errors.append(result.error)
        result = self.create_all_folderpaths()
        if result.failure:
            errors.append(result.error)
        if not errors:
            self.start_time = datetime.now()
            return Result.Ok()
        return Result.Fail("\n".join(errors))

    def show_status_report(self):
        if self.status_report == StatusReport.SEASON_SUMMARY:
            return report_season_status(
                db_session=self.db_session,
                year=self.season.year,
                report_type=self.status_report,
            )
        return report_date_range_status(
            db_session=self.db_session,
            start_date=self.db_job.start_date,
            end_date=self.db_job.end_date,
            report_type=self.status_report,
        )

    def job_failed(self, result, data_set=DataSet.ALL):
        self.end_time = datetime.now()
        self.db_job.status = JobStatus.ERROR
        new_error = ScrapeError(
            error_message=result.error, data_set=data_set, job_id=self.db_job.id
        )
        self.db_session.add(new_error)
        self.db_session.commit()
        return result

    def job_succeeded(self):
        self.end_time = datetime.now()
        self.db_job.status = JobStatus.COMPLETE
        self.db_session.commit()
        shutil.rmtree(self.db_job.scraped_html_root_folder)
        return Result.Ok()

    def create_all_folderpaths(self):
        return self.scraped_data.create_all_folderpaths(self.season.year)

    def check_url_delay_settings(self):
        return self.config.check_url_delay_settings(self.data_sets)

    def s3_bucket_required(self):
        return self.config.s3_bucket_required(self.data_sets)

    def check_s3_bucket(self):
        return self.scraped_data.check_s3_bucket()



"""Weather conditions at game start time and other descriptive info."""
from dataclasses import dataclass


@dataclass
class BBRefBoxscoreMeta:
    """Weather conditions at game start time and other descriptive info."""

    park_name: str = ""
    field_type: str = ""
    day_night: str = ""
    first_pitch_temperature: int = 0
    first_pitch_precipitation: str = ""
    first_pitch_wind: str = ""
    first_pitch_clouds: str = ""
    game_duration: str = ""
    attendance: int = 0

    def as_dict(self):
        """Convert boxscore meta info to a dictionary."""
        return dict(
            __bbref_boxscore_meta__=True,
            park_name=self.park_name,
            field_type=self.field_type,
            day_night=self.day_night,
            first_pitch_temperature=self.first_pitch_temperature,
            first_pitch_precipitation=self.first_pitch_precipitation,
            first_pitch_wind=self.first_pitch_wind,
            first_pitch_clouds=self.first_pitch_clouds,
            game_duration=self.game_duration,
            attendance=self.attendance,
        )



from dataclasses import dataclass
from datetime import datetime

from dataclass_csv import accept_whitespaces, dateformat
from sqlalchemy import Column, DateTime, ForeignKey, Integer
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

from vigorish.config.database import Base
from vigorish.util.dataclass_helpers import dict_from_dataclass, sanitize_row_dict
from vigorish.util.dt_format_strings import DATE_ONLY, DATE_ONLY_2, DATE_ONLY_TABLE_ID


class DateScrapeStatus(Base):

    __tablename__ = "scrape_status_date"
    id = Column(Integer, primary_key=True)
    game_date = Column(DateTime)
    scraped_daily_dash_bbref = Column(Integer, default=0)
    scraped_daily_dash_brooks = Column(Integer, default=0)
    game_count_bbref = Column(Integer, default=0)
    game_count_brooks = Column(Integer, default=0)
    season_id = Column(Integer, ForeignKey("season.id"))

    scrape_status_games = relationship("GameScrapeStatus", backref="scrape_status_date")
    scrape_status_pitchfx = relationship("PitchAppScrapeStatus", backref="scrape_status_date")
    pitch_app_status = relationship(
        "Date_PitchApp_View",
        backref="original",
        uselist=False,
        primaryjoin="DateScrapeStatus.id==Date_PitchApp_View.id",
        foreign_keys="Date_PitchApp_View.id",
    )

    @hybrid_property
    def game_date_str(self):
        return self.game_date.strftime(DATE_ONLY_2)

    @hybrid_property
    def total_games(self):
        return len(self.scrape_status_games) if self.scrape_status_games else 0

    @hybrid_property
    def total_games_combined_success(self):
        return sum(game.combined_data_success for game in self.scrape_status_games)

    @hybrid_property
    def total_games_combined_fail(self):
        return sum(game.combined_data_fail for game in self.scrape_status_games)

    @hybrid_property
    def total_games_combined(self):
        return self.total_games_combined_success + self.total_games_combined_fail

    @hybrid_property
    def total_bbref_boxscores_scraped(self):
        return sum(game.scraped_bbref_boxscore for game in self.scrape_status_games)

    @hybrid_property
    def percent_complete_bbref_boxscores_scraped(self):
        return (
            self.total_bbref_boxscores_scraped / float(len(self.scrape_status_games))
            if len(self.scrape_status_games) > 0
            else 0.0
        )

    @hybrid_property
    def scraped_all_bbref_boxscores(self):
        if not self.scraped_daily_dash_bbref or not self.scraped_daily_dash_brooks:
            return False
        return all(game.scraped_bbref_boxscore == 1 for game in self.scrape_status_games)

    @hybrid_property
    def total_brooks_pitch_logs_scraped(self):
        return sum(game.scraped_brooks_pitch_logs for game in self.scrape_status_games)

    @hybrid_property
    def percent_complete_brooks_pitch_logs(self):
        return (
            self.total_brooks_pitch_logs_scraped / float(len(self.scrape_status_games))
            if len(self.scrape_status_games) > 0
            else 0.0
        )

    @hybrid_property
    def scraped_all_brooks_pitch_logs(self):
        if not self.scraped_daily_dash_bbref or not self.scraped_daily_dash_brooks:
            return False
        return all(game.scraped_brooks_pitch_logs == 1 for game in self.scrape_status_games)

    @hybrid_property
    def pitch_app_count_bbref(self):
        return sum(game.pitch_app_count_bbref for game in self.scrape_status_games)

    @hybrid_property
    def pitch_app_count_brooks(self):
        return sum(game.pitch_app_count_brooks for game in self.scrape_status_games)

    @hybrid_property
    def total_pitch_count_bbref(self):
        return sum(game.total_pitch_count_bbref for game in self.scrape_status_games)

    @hybrid_property
    def pitch_app_count_pitchfx(self):
        return self.pitch_app_status.total_pitchfx if self.pitch_app_status.total_pitchfx else 0

    @hybrid_property
    def total_pitch_apps_scraped_pitchfx(self):
        return (
            self.pitch_app_status.total_pitchfx_scraped
            if self.pitch_app_status.total_pitchfx_scraped
            else 0
        )

    @hybrid_property
    def total_pitch_apps_no_pitchfx_data(self):
        return (
            self.pitch_app_status.total_no_pitchfx_data
            if self.pitch_app_status.total_no_pitchfx_data
            else 0
        )

    @hybrid_property
    def total_pitch_apps_with_pitchfx_data(self):
        return self.pitch_app_count_pitchfx - self.total_pitch_apps_no_pitchfx_data

    @hybrid_property
    def total_pitch_apps_combined_data(self):
        return (
            self.pitch_app_status.total_combined_pitchfx_bbref_data
            if self.pitch_app_status.total_combined_pitchfx_bbref_data
            else 0
        )

    @hybrid_property
    def total_pitch_apps_pitchfx_error(self):
        return (
            self.pitch_app_status.total_pitchfx_error
            if self.pitch_app_status.total_pitchfx_error
            else 0
        )

    @hybrid_property
    def total_pitch_apps_invalid_pitchfx(self):
        return (
            self.pitch_app_status.total_invalid_pitchfx
            if self.pitch_app_status.total_invalid_pitchfx
            else 0
        )

    @hybrid_property
    def total_pitch_apps_any_pitchfx_error(self):
        return +self.total_pitch_apps_pitchfx_error + self.total_pitch_apps_invalid_pitchfx

    @hybrid_property
    def total_pitch_apps_pitchfx_is_valid(self):
        return sum(
            not (pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.scrape_status_pitchfx
        )

    @hybrid_property
    def total_pitch_count_pitch_logs(self):
        return (
            self.pitch_app_status.total_pitch_count_pitch_log
            if self.pitch_app_status.total_pitch_count_pitch_log
            else 0
        )

    @hybrid_property
    def total_pitch_count_bbref_audited(self):
        return (
            self.pitch_app_status.total_pitch_count_bbref
            if self.pitch_app_status.total_pitch_count_bbref
            else 0
        )

    @hybrid_property
    def total_pitch_count_pitchfx(self):
        return (
            self.pitch_app_status.total_pitch_count_pitchfx
            if self.pitch_app_status.total_pitch_count_pitchfx
            else 0
        )

    @hybrid_property
    def total_pitch_count_pitchfx_audited(self):
        return (
            self.pitch_app_status.total_pitch_count_pitchfx_audited
            if self.pitch_app_status.total_pitch_count_pitchfx_audited
            else 0
        )

    @hybrid_property
    def total_duplicate_pitchfx_removed_count(self):
        return (
            self.pitch_app_status.total_duplicate_pitchfx_removed_count
            if self.pitch_app_status.total_duplicate_pitchfx_removed_count
            else 0
        )

    @hybrid_property
    def total_missing_pitchfx_count(self):
        return (
            self.pitch_app_status.total_missing_pitchfx_count
            if self.pitch_app_status.total_missing_pitchfx_count
            else 0
        )

    @hybrid_property
    def total_extra_pitchfx_count(self):
        return (
            self.pitch_app_status.total_extra_pitchfx_count
            if self.pitch_app_status.total_extra_pitchfx_count
            else 0
        )

    @hybrid_property
    def total_extra_pitchfx_removed_count(self):
        return (
            self.pitch_app_status.total_extra_pitchfx_removed_count
            if self.pitch_app_status.total_extra_pitchfx_removed_count
            else 0
        )

    @hybrid_property
    def total_batters_faced_bbref(self):
        return (
            self.pitch_app_status.total_batters_faced_bbref
            if self.pitch_app_status.total_batters_faced_bbref
            else 0
        )

    @hybrid_property
    def total_batters_faced_pitchfx(self):
        return (
            self.pitch_app_status.total_batters_faced_pitchfx
            if self.pitch_app_status.total_batters_faced_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_pitchfx_complete(self):
        return (
            self.pitch_app_status.total_at_bats_pitchfx_complete
            if self.pitch_app_status.total_at_bats_pitchfx_complete
            else 0
        )

    @hybrid_property
    def total_at_bats_missing_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_missing_pitchfx
            if self.pitch_app_status.total_at_bats_missing_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_extra_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_extra_pitchfx
            if self.pitch_app_status.total_at_bats_extra_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_extra_pitchfx_removed(self):
        return (
            self.pitch_app_status.total_at_bats_extra_pitchfx_removed
            if self.pitch_app_status.total_at_bats_pitchfx_error
            else 0
        )

    @hybrid_property
    def total_at_bats_pitchfx_error(self):
        return (
            self.pitch_app_status.total_at_bats_pitchfx_error
            if self.pitch_app_status.total_at_bats_pitchfx_error
            else 0
        )

    @hybrid_property
    def total_at_bats_invalid_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_invalid_pitchfx
            if self.pitch_app_status.total_at_bats_invalid_pitchfx
            else 0
        )

    @hybrid_property
    def scraped_all_pitchfx_logs(self):
        if not self.scraped_all_brooks_pitch_logs:
            return False
        if not self.scrape_status_pitchfx:
            return True
        return self.pitch_app_count_pitchfx == self.total_pitch_apps_scraped_pitchfx

    @hybrid_property
    def combined_data_for_all_pitchfx_logs(self):
        if not self.scraped_all_pitchfx_logs:
            return False
        if not self.scrape_status_pitchfx:
            return False
        return self.pitch_app_count_pitchfx == self.total_pitch_apps_combined_data

    @hybrid_property
    def pitchfx_error_for_any_pitchfx_logs(self):
        if not self.scraped_all_pitchfx_logs:
            return False
        if not self.scrape_status_pitchfx:
            return False
        return any((pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.scrape_status_pitchfx)

    @hybrid_property
    def pitchfx_is_valid_for_all_pitchfx_logs(self):
        if not self.scraped_all_pitchfx_logs:
            return True
        if not self.scrape_status_pitchfx:
            return True
        return all(
            not (pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.scrape_status_pitchfx
        )

    @hybrid_property
    def percent_complete_pitchfx_logs_scraped(self):
        return (
            self.total_pitch_apps_scraped_pitchfx / float(self.pitch_app_count_pitchfx)
            if self.pitch_app_count_pitchfx > 0
            else 0.0
        )

    @hybrid_property
    def scraped_no_data(self):
        return self.scraped_daily_dash_bbref == 0 and self.scraped_daily_dash_brooks == 0

    @hybrid_property
    def scraped_only_bbref_daily_dash(self):
        return (
            self.scraped_daily_dash_bbref == 1
            and self.scraped_daily_dash_brooks == 0
            and not self.scraped_all_bbref_boxscores
            and not self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_pitchfx_logs
        )

    @hybrid_property
    def scraped_only_brooks_daily_dash(self):
        return (
            self.scraped_daily_dash_bbref == 0
            and self.scraped_daily_dash_brooks == 1
            and not self.scraped_all_bbref_boxscores
            and not self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_pitchfx_logs
        )

    @hybrid_property
    def scraped_only_both_daily_dash(self):
        return (
            self.scraped_daily_dash_bbref == 1
            and self.scraped_daily_dash_brooks == 1
            and not self.scraped_all_bbref_boxscores
            and not self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_pitchfx_logs
        )

    @hybrid_property
    def scraped_only_bbref_boxscores(self):
        if (
            self.scraped_all_bbref_boxscores
            and not self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_pitchfx_logs
        ):
            return True
        return False

    @hybrid_property
    def scraped_only_brooks_pitch_logs(self):
        if (
            self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_bbref_boxscores
            and not self.scraped_all_pitchfx_logs
        ):
            return True
        return False

    @hybrid_property
    def scraped_only_both_bbref_boxscores_and_brooks_pitch_logs(self):
        if (
            self.scraped_all_bbref_boxscores
            and self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_pitchfx_logs
        ):
            return True
        return False

    @hybrid_property
    def scraped_all_game_data(self):
        return (
            self.scraped_all_bbref_boxscores
            and self.scraped_all_brooks_pitch_logs
            and self.scraped_all_pitchfx_logs
        )

    @hybrid_property
    def scrape_status_description(self):
        if self.scraped_all_game_data:
            return "Scraped all game data"
        elif self.scraped_only_both_bbref_boxscores_and_brooks_pitch_logs:
            return (
                "Missing Brooks pitchfx logs "
                f"({self.total_pitch_apps_scraped_pitchfx}/{self.pitch_app_count_brooks}, "
                f"{self.percent_complete_pitchfx_logs_scraped:.0%})"
            )
        elif self.scraped_only_brooks_pitch_logs:
            return "Scraped Brooks pitch logs, missing BBref boxscores"
        elif self.scraped_only_bbref_boxscores:
            return "Scraped BBref boxscores, missing Brooks pitch logs"
        elif self.scraped_only_both_daily_dash:
            return "Missing BBref boxscores and Brooks pitch logs"
        elif self.scraped_only_brooks_daily_dash:
            return "Scraped Brooks daily dashboard, missing BBref daily dash"
        elif self.scraped_only_bbref_daily_dash:
            return "Scraped BBref daily dashboard, missing Brooks daily dash"
        elif self.scraped_no_data:
            return "No data has been scraped"
        else:
            return "N/A"

    def __init__(self, **kwargs):
        super(DateScrapeStatus, self).__init__(**kwargs)
        self.id = self.game_date.strftime(DATE_ONLY_TABLE_ID)

    def __repr__(self):
        return f"<DateScrapeStatus date={self.game_date_str}, season_id={self.season_id}>"

    def as_dict(self):
        d = {c.name: getattr(self, c.name) for c in self.__table__.columns}
        d["game_date_str"] = self.game_date_str
        d["total_games"] = self.total_games
        d["total_games_combined_success"] = self.total_games_combined_success
        d["total_games_combined_fail"] = self.total_games_combined_fail
        d["total_games_combined"] = self.total_games_combined
        d["total_bbref_boxscores_scraped"] = self.total_bbref_boxscores_scraped
        d[
            "percent_complete_bbref_boxscores_scraped"
        ] = f"{self.percent_complete_bbref_boxscores_scraped:01.0f}%"
        d["scraped_all_bbref_boxscores"] = self.scraped_all_bbref_boxscores
        d["total_brooks_pitch_logs_scraped"] = self.total_brooks_pitch_logs_scraped
        d[
            "percent_complete_brooks_pitch_logs"
        ] = f"{self.percent_complete_brooks_pitch_logs:01.0f}%"
        d["scraped_all_brooks_pitch_logs"] = self.scraped_all_brooks_pitch_logs
        d["pitch_app_count_bbref"] = self.pitch_app_count_bbref
        d["pitch_app_count_brooks"] = self.pitch_app_count_brooks
        d["total_pitch_count_bbref"] = self.total_pitch_count_bbref
        d["pitch_app_count_pitchfx"] = self.pitch_app_count_pitchfx
        d["total_pitch_apps_scraped_pitchfx"] = self.total_pitch_apps_scraped_pitchfx
        d["total_pitch_apps_no_pitchfx_data"] = self.total_pitch_apps_no_pitchfx_data
        d["total_pitch_apps_with_pitchfx_data"] = self.total_pitch_apps_with_pitchfx_data
        d["total_pitch_apps_combined_data"] = self.total_pitch_apps_combined_data
        d["total_pitch_apps_pitchfx_error"] = self.total_pitch_apps_pitchfx_error
        d["total_pitch_apps_invalid_pitchfx"] = self.total_pitch_apps_invalid_pitchfx
        d["total_pitch_apps_pitchfx_is_valid"] = self.total_pitch_apps_pitchfx_is_valid
        d["total_pitch_count_pitch_logs"] = self.total_pitch_count_pitch_logs
        d["total_pitch_count_bbref_audited"] = self.total_pitch_count_bbref_audited
        d["total_pitch_count_pitchfx"] = self.total_pitch_count_pitchfx
        d["total_pitch_count_pitchfx_audited"] = self.total_pitch_count_pitchfx_audited
        d["total_duplicate_pitchfx_removed_count"] = self.total_duplicate_pitchfx_removed_count
        d["total_missing_pitchfx_count"] = self.total_missing_pitchfx_count
        d["total_extra_pitchfx_count"] = self.total_extra_pitchfx_count
        d["total_extra_pitchfx_removed_count"] = self.total_extra_pitchfx_removed_count
        d["total_batters_faced_bbref"] = self.total_batters_faced_bbref
        d["total_batters_faced_pitchfx"] = self.total_batters_faced_pitchfx
        d["total_at_bats_pitchfx_complete"] = self.total_at_bats_pitchfx_complete
        d["total_at_bats_missing_pitchfx"] = self.total_at_bats_missing_pitchfx
        d["total_at_bats_extra_pitchfx"] = self.total_at_bats_extra_pitchfx
        d["total_at_bats_extra_pitchfx_removed"] = self.total_at_bats_extra_pitchfx_removed
        d["total_at_bats_pitchfx_error"] = self.total_at_bats_pitchfx_error
        d["total_at_bats_invalid_pitchfx"] = self.total_at_bats_invalid_pitchfx
        d["scraped_all_pitchfx_logs"] = self.scraped_all_pitchfx_logs
        d["combined_data_for_all_pitchfx_logs"] = self.combined_data_for_all_pitchfx_logs
        d["pitchfx_error_for_any_pitchfx_logs"] = self.pitchfx_error_for_any_pitchfx_logs
        d["pitchfx_is_valid_for_all_pitchfx_logs"] = self.pitchfx_is_valid_for_all_pitchfx_logs
        d["percent_complete_pitchfx_logs_scraped"] = self.percent_complete_pitchfx_logs_scraped
        d["scraped_no_data"] = self.scraped_only_bbref_daily_dash
        d["scraped_only_bbref_daily_dash"] = self.scraped_only_bbref_daily_dash
        d["scraped_only_brooks_daily_dash"] = self.scraped_only_brooks_daily_dash
        d["scraped_only_both_daily_dash"] = self.scraped_only_both_daily_dash
        d["scraped_only_bbref_boxscores"] = self.scraped_only_bbref_boxscores
        d["scraped_only_brooks_pitch_logs"] = self.scraped_only_brooks_pitch_logs
        d[
            "scraped_only_both_bbref_boxscores_and_brooks_pitch_logs"
        ] = self.scraped_only_both_bbref_boxscores_and_brooks_pitch_logs
        d["scraped_all_game_data"] = self.scraped_all_game_data
        d["scrape_status_description"] = self.scrape_status_description
        return d

    def status_report(self):
        scraped_daily_bbref = "YES" if self.scraped_daily_dash_bbref == 1 else "NO"
        scraped_daily_brooks = "YES" if self.scraped_daily_dash_brooks == 1 else "NO"
        scraped_bbref_boxscores = "YES" if self.scraped_all_bbref_boxscores else "NO"
        scraped_brooks_pitch_logs = "YES" if self.scraped_all_brooks_pitch_logs else "NO"
        scraped_all_pitchfx_logs = "YES" if self.scraped_all_pitchfx_logs else "NO"
        combined_data_for_all_pitchfx_logs = (
            "YES" if self.combined_data_for_all_pitchfx_logs else "NO"
        )
        pitchfx_error_for_any_pitchfx_logs = (
            "YES" if self.pitchfx_error_for_any_pitchfx_logs else "NO"
        )
        total_pitchfx_removed_count = (
            self.total_duplicate_pitchfx_removed_count + self.total_extra_pitchfx_removed_count
        )
        return (
            f"Overall Status For Date......................: {self.scrape_status_description}\n"
            f"Scraped Daily Dashboard (BBRef/Brooks).......: "
            f"{scraped_daily_bbref}/{scraped_daily_brooks}\n"
            f"BBref Boxscores Scraped......................: {scraped_bbref_boxscores} "
            f"{self.total_bbref_boxscores_scraped}/{self.total_games}\n"
            f"Brooks Games Scraped.........................: {scraped_brooks_pitch_logs} "
            f"{self.total_brooks_pitch_logs_scraped}/{self.total_games}\n"
            f"PitchFx Logs Scraped.........................: {scraped_all_pitchfx_logs} "
            f"{self.total_pitch_apps_scraped_pitchfx}/{self.pitch_app_count_pitchfx} "
            f"({self.percent_complete_pitchfx_logs_scraped:.0%})\n"
            f"Combined BBRef/PitchFX Data (Success/Total)..: {combined_data_for_all_pitchfx_logs} "
            f"{self.total_games_combined_success}/{self.total_games_combined}\n"
            f"Pitch App Count (BBRef/Brooks)...............: "
            f"{self.pitch_app_count_bbref}/{self.pitch_app_count_brooks}\n"
            f"Pitch App Count (PFx/data/no data)...........: {self.pitch_app_count_pitchfx}/"
            f"{self.total_pitch_apps_with_pitchfx_data}/{self.total_pitch_apps_no_pitchfx_data}\n"
            f"PitchFX Data Errors (Valid AB/Invalid AB)....: {pitchfx_error_for_any_pitchfx_logs} "
            f"{self.total_pitch_apps_pitchfx_error}/{self.total_pitch_apps_invalid_pitchfx}\n"
            f"Pitch Count (BBRef/Brooks/PFx)...............: {self.total_pitch_count_bbref}/"
            f"{self.total_pitch_count_pitch_logs}/{self.total_pitch_count_pitchfx}\n"
            "Pitch Count Audited (BBRef/PFx/Removed)......: "
            f"{self.total_pitch_count_bbref_audited}/{self.total_pitch_count_pitchfx_audited}/"
            f"{total_pitchfx_removed_count}\n"
        )

    def games_status_report(self):
        return "\n".join([game_status.status_report() for game_status in self.scrape_status_games])

    def as_csv_dict(self):
        return dict_from_dataclass(self, DateScrapeStatusCsvRow, date_format=DATE_ONLY)

    @classmethod
    def create_new(cls, game_date, season_id):
        game_date_str = game_date.strftime(DATE_ONLY_TABLE_ID)
        return cls(id=int(game_date_str), game_date=game_date, season_id=season_id)

    @classmethod
    def get_csv_col_names(cls):
        return [name for name in DateScrapeStatusCsvRow.__dataclass_fields__.keys()]

    @classmethod
    def export_table_as_csv(cls, db_session):
        col_names = ",".join(cls.get_csv_col_names())
        csv_dicts = (obj.as_csv_dict() for obj in db_session.query(cls).all())
        csv_rows = (",".join(sanitize_row_dict(d, date_format=DATE_ONLY)) for d in csv_dicts)
        yield col_names
        for row in csv_rows:
            yield row

    @classmethod
    def find_by_date(cls, db_session, game_date):
        date_str = game_date.strftime(DATE_ONLY_TABLE_ID)
        return db_session.query(cls).get(int(date_str))

    @classmethod
    def get_all_bbref_scraped_dates_for_season(cls, db_session, season_id):
        return [
            date_status.game_date
            for date_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if date_status.scraped_daily_dash_bbref == 1
        ]

    @classmethod
    def get_all_bbref_unscraped_dates_for_season(cls, db_session, season_id):
        return [
            date_status.game_date
            for date_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if date_status.scraped_daily_dash_bbref == 0
        ]

    @classmethod
    def get_all_brooks_scraped_dates_for_season(cls, db_session, season_id):
        return [
            date_status.game_date
            for date_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if date_status.scraped_daily_dash_brooks == 1
        ]

    @classmethod
    def get_all_brooks_unscraped_dates_for_season(cls, db_session, season_id):
        return [
            date_status.game_date
            for date_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if date_status.scraped_daily_dash_brooks == 0
        ]

    @classmethod
    def get_unscraped_pitch_appearances_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        return [
            pitch_app
            for pitch_app in date_status.scrape_status_pitchfx
            if pitch_app.scraped_pitchfx == 0
        ]

    @classmethod
    def get_unscraped_pitch_app_ids_for_date(cls, db_session, game_date):
        unscraped_pitch_apps = cls.get_unscraped_pitch_appearances_for_date(db_session, game_date)
        return [pitch_app.pitch_app_id for pitch_app in unscraped_pitch_apps]

    @classmethod
    def verify_bbref_daily_dashboard_scraped_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        return date_status.scraped_daily_dash_bbref == 1 if date_status else False

    @classmethod
    def verify_brooks_daily_dashboard_scraped_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        return date_status.scraped_daily_dash_brooks == 1 if date_status else False

    @classmethod
    def verify_all_bbref_boxscores_scraped_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        return date_status.scraped_all_bbref_boxscores if date_status else False

    @classmethod
    def verify_all_brooks_pitch_logs_scraped_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        return date_status.scraped_all_brooks_pitch_logs if date_status else False

    @classmethod
    def verify_all_brooks_pitchfx_scraped_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        return date_status.scraped_all_pitchfx_logs if date_status else False

    @classmethod
    def get_all_bbref_game_ids_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        if not date_status:
            return None
        return [game_status.bbref_game_id for game_status in date_status.scrape_status_games]

    @classmethod
    def get_all_brooks_game_ids_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        if not date_status:
            return None
        return [game_status.bb_game_id for game_status in date_status.scrape_status_games]

    @classmethod
    def get_date_id_map(cls, db_session):
        all_dates = db_session.query(cls).all()
        return {d.game_date: d.id for d in all_dates}


@accept_whitespaces
@dateformat(DATE_ONLY)
@dataclass
class DateScrapeStatusCsvRow:
    id: int
    game_date: datetime = None
    scraped_daily_dash_bbref: int = 0
    scraped_daily_dash_brooks: int = 0
    game_count_bbref: int = 0
    game_count_brooks: int = 0



from sqlalchemy import func, join, select
from sqlalchemy_utils import create_view

from vigorish.config.database import Base, DateScrapeStatus, Season


class Season_Date_View(Base):
    __table__ = create_view(
        name="season_date_status",
        selectable=select(
            [
                Season.id.label("id"),
                func.count(DateScrapeStatus.id).label("total_days"),
                func.sum(DateScrapeStatus.scraped_daily_dash_bbref).label(
                    "total_scraped_daily_dash_bbref"
                ),
                func.sum(DateScrapeStatus.scraped_daily_dash_brooks).label(
                    "total_scraped_daily_dash_brooks"
                ),
                func.sum(DateScrapeStatus.game_count_bbref).label("total_game_count_bbref"),
                func.sum(DateScrapeStatus.game_count_brooks).label("total_game_count_brooks"),
            ]
        )
        .select_from(
            join(
                Season,
                DateScrapeStatus,
                Season.id == DateScrapeStatus.season_id,
                isouter=True,
            )
        )
        .group_by(Season.id),
        metadata=Base.metadata,
        cascade_on_drop=False,
    )



"""Menu item that returns the user to the previous menu."""
import subprocess

from getch import pause
from halo import Halo

from vigorish.cli.components import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_heading,
    print_message,
)
from vigorish.enums import DataSet, SyncDirection, VigFile
from vigorish.tasks.base import Task
from vigorish.tasks.sync_scraped_data import SyncScrapedData
from vigorish.util.result import Result

SYNC_STATUS_TEXT_COLOR = {
    "out_of_sync": "bright_green",
    "in_sync": "blue",
    "sync_complete": "bright_green",
    "error": "bright_red",
}


class SyncScrapedDataNoPrompts(Task):
    def __init__(self, app):
        super().__init__(app)
        self.sync_direction = None
        self.year = None
        self.file_type = None
        self.data_sets_int = 0
        self.current_data_set = None
        self.task_number = 0
        self.sync_files = {}
        self.sync_results = []
        self.task_number = 0
        self.spinners = {}
        self.results = {}

    @property
    def bbref_games_for_date(self):
        return self.data_sets_int & DataSet.BBREF_GAMES_FOR_DATE == DataSet.BBREF_GAMES_FOR_DATE

    @property
    def bbref_boxscores(self):
        return self.data_sets_int & DataSet.BBREF_BOXSCORES == DataSet.BBREF_BOXSCORES

    @property
    def brooks_games_for_date(self):
        return self.data_sets_int & DataSet.BROOKS_GAMES_FOR_DATE == DataSet.BROOKS_GAMES_FOR_DATE

    @property
    def brooks_pitch_logs(self):
        return self.data_sets_int & DataSet.BROOKS_PITCH_LOGS == DataSet.BROOKS_PITCH_LOGS

    @property
    def brooks_pitchfx(self):
        return self.data_sets_int & DataSet.BROOKS_PITCHFX == DataSet.BROOKS_PITCHFX

    @property
    def data_sets(self):
        if self.file_type == VigFile.COMBINED_GAME_DATA:
            return [DataSet.ALL]
        data_sets = []
        if self.brooks_games_for_date:
            data_sets.append(DataSet.BROOKS_GAMES_FOR_DATE)
        if self.brooks_pitch_logs:
            data_sets.append(DataSet.BROOKS_PITCH_LOGS)
        if self.brooks_pitchfx:
            data_sets.append(DataSet.BROOKS_PITCHFX)
        if self.bbref_games_for_date:
            data_sets.append(DataSet.BBREF_GAMES_FOR_DATE)
        if self.bbref_boxscores:
            data_sets.append(DataSet.BBREF_BOXSCORES)
        return sorted(ds for ds in data_sets if ds in self.valid_data_sets())

    @property
    def total_tasks(self):
        return len(self.data_sets)

    @property
    def all_files_are_in_sync(self):
        return all(not out_of_sync for (out_of_sync, _, _) in self.sync_files.values())

    def execute(self, sync_direction, year, file_type, data_sets_int):
        self.sync_direction = sync_direction
        self.year = year
        self.file_type = file_type
        self.data_sets_int = data_sets_int
        self.initialize_s3_sync_task()
        for data_set in self.data_sets:
            self.task_number += 1
            self.current_data_set = data_set
            self.report_sync_results()
            result = self.s3_sync.execute(sync_direction, file_type, data_set, year)
            self.results[data_set] = result
            self.spinners[data_set].stop()
            if result.failure:
                return self.results
        self.report_sync_results()
        self.teardown()
        pause(message="\nPress any key to continue...")
        return self.results

    def initialize_s3_sync_task(self):
        s3_objects = self.get_all_objects_in_s3()
        self.s3_sync = SyncScrapedData(self.app, cached_s3_objects=s3_objects)
        self.s3_sync.events.error_occurred += self.error_occurred
        self.s3_sync.events.get_sync_files_start += self.get_sync_files_start
        self.s3_sync.events.get_sync_files_complete += self.get_sync_files_complete
        self.s3_sync.events.sync_files_start += self.sync_files_start
        self.s3_sync.events.sync_files_progress += self.sync_files_progress
        self.s3_sync.events.sync_files_complete += self.sync_files_complete

    def get_all_objects_in_s3(self):
        subprocess.run(["clear"])
        spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        spinner.text = "Retrieving data for all objects stored in S3..."
        spinner.start()
        s3_obj_collection = self.scraped_data.file_helper.get_all_object_keys_in_s3_bucket()
        s3_objects = [obj for obj in s3_obj_collection]
        spinner.stop()
        return s3_objects

    def valid_data_sets(self):
        data_set_file_type_map = {
            VigFile.SCRAPED_HTML: [
                DataSet.BBREF_GAMES_FOR_DATE,
                DataSet.BROOKS_GAMES_FOR_DATE,
                DataSet.BBREF_BOXSCORES,
                DataSet.BROOKS_PITCH_LOGS,
                DataSet.BROOKS_PITCHFX,
            ],
            VigFile.PARSED_JSON: [
                DataSet.BBREF_GAMES_FOR_DATE,
                DataSet.BROOKS_GAMES_FOR_DATE,
                DataSet.BBREF_BOXSCORES,
                DataSet.BROOKS_PITCH_LOGS,
                DataSet.BROOKS_PITCHFX,
            ],
            VigFile.COMBINED_GAME_DATA: [DataSet.ALL],
            VigFile.PATCH_LIST: [DataSet.BBREF_BOXSCORES, DataSet.BROOKS_PITCHFX],
        }
        return data_set_file_type_map[self.file_type]

    def report_sync_results(self):
        subprocess.run(["clear"])
        self.print_header_message()
        if not self.sync_results:
            return
        for task_result in self.sync_results:
            print_message(task_result[0], wrap=False, fg=task_result[1])

    def print_header_message(self):
        if self.sync_direction == SyncDirection.UP_TO_S3:
            heading = "Syncing data from local folder to S3 bucket"
        if self.sync_direction == SyncDirection.DOWN_TO_LOCAL:
            heading = "Syncing data from S3 bucket to local folder"
        print_heading(heading, fg="bright_yellow")

    def teardown(self):
        self.s3_sync.events.error_occurred -= self.error_occurred
        self.s3_sync.events.get_sync_files_start -= self.get_sync_files_start
        self.s3_sync.events.get_sync_files_complete -= self.get_sync_files_complete
        self.s3_sync.events.sync_files_start -= self.sync_files_start
        self.s3_sync.events.sync_files_progress -= self.sync_files_progress
        self.s3_sync.events.sync_files_complete -= self.sync_files_complete

    def error_occurred(self, error_message):
        self.sync_results.append((error_message, SYNC_STATUS_TEXT_COLOR["error"]))
        self.results[self.current_data_set] = Result.Fail(error_message)

    def get_sync_files_start(self):
        self.spinners[self.current_data_set] = Halo(
            spinner=get_random_dots_spinner(),
            color=get_random_cli_color(),
            text=self.sync_started_text(),
        )
        self.spinners[self.current_data_set].start()

    def sync_started_text(self):
        return (
            f"Analyzing MLB {self.year} {self.file_type} {self.current_data_set} files "
            f"(Task {self.task_number}/{self.total_tasks})..."
        )

    def get_sync_files_complete(self, sync_results):
        (out_of_sync, new_files, update_files) = sync_results
        self.sync_files[self.current_data_set] = (out_of_sync, new_files, update_files)
        if out_of_sync:
            sync_count = len(new_files) + len(update_files)
            self.sync_results.append(self.out_of_sync_text(sync_count))
        else:
            self.sync_results.append(self.in_sync_text())

    def out_of_sync_text(self, sync_count):
        out_of_sync = (
            f"[{self.year} {self.file_type} {self.current_data_set}] "
            f"{sync_count} files out of sync! "
            f"(Task {self.task_number}/{self.total_tasks})"
        )
        return (out_of_sync, SYNC_STATUS_TEXT_COLOR["out_of_sync"])

    def in_sync_text(self):
        in_sync = (
            f"[{self.year} {self.file_type} {self.current_data_set}] "
            f"All files in sync "
            f"(Task {self.task_number}/{self.total_tasks})"
        )
        return (in_sync, SYNC_STATUS_TEXT_COLOR["in_sync"])

    def sync_files_start(self, name, complete, total):
        self.report_sync_results()
        self.sync_files_progress(name, complete, total)

    def sync_files_progress(self, name, complete, total):
        direction = "Down" if self.sync_direction == "SYNC_DOWN" else "Up"
        percent = complete / float(total)
        progress_message = f"{direction}loading: {name} | {percent:.0%} ({complete}/{total} Files)"
        self.spinners[self.current_data_set].text = progress_message

    def sync_files_complete(self):
        if self.sync_direction == SyncDirection.UP_TO_S3:
            sync_complete = (
                f"[{self.year} {self.file_type} {self.current_data_set}] "
                f"All changes have been synced to s3 bucket!"
            )
        if self.sync_direction == SyncDirection.DOWN_TO_LOCAL:
            sync_complete = (
                f"[{self.year} {self.file_type} {self.current_data_set}] "
                f"All changes have been synced to local folder!"
            )
        self.sync_results.append((sync_complete, SYNC_STATUS_TEXT_COLOR["sync_complete"]))



"""Menu item that provides detailed information for a single job."""
import subprocess

from bullet import Bullet, colors
from getch import pause

from vigorish.cli.components import print_message
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJI_DICT, MENU_NUMBERS
from vigorish.enums import JobStatus
from vigorish.scrape.job_runner import JobRunner
from vigorish.util.list_helpers import report_dict
from vigorish.util.result import Result


class JobDetails(MenuItem):
    def __init__(self, app, db_job, menu_item_number):
        super().__init__(app)
        self.db_job = db_job
        self.job_status = db_job.status
        self.job_details = db_job.job_details
        self.menu_item_text = f" {db_job.name} (Status: {db_job.status.name}, ID: {db_job.id})"
        self.menu_item_emoji = MENU_NUMBERS.get(menu_item_number, f"{menu_item_number}. ")
        self.exit_menu = False

    def launch(self):
        subprocess.run(["clear"])
        print_message("*** Job Details ***", fg="bright_yellow", bold=True)
        job_details = report_dict(self.job_details, title="", title_prefix="", title_suffix="")
        print_message(f"{job_details}\n", wrap=False, fg="bright_yellow")
        if self.db_job.errors:
            print_message("*** Errors ***", fg="bright_red", bold=True)
            print_message(self.db_job.error_messages, fg="bright_red")
        if self.job_status == JobStatus.INCOMPLETE:
            result = self.incomplete_job_options_prompt()
            if result.failure:
                return Result.Ok(self.exit_menu)
            user_choice = result.value
            if user_choice == "RUN":
                job_runner = JobRunner(
                    db_job=self.db_job,
                    db_session=self.db_session,
                    config=self.config,
                    scraped_data=self.scraped_data,
                )
                result = job_runner.execute()
                if result.failure:
                    return result
                return Result.Ok(True)
            if user_choice == "CANCEL":
                self.db_job.status = JobStatus.COMPLETE
                self.db_session.commit()
                subprocess.run(["clear"])
                print_message("Job was successfully cancelled.", fg="bright_red", bold=True)
                pause(message="Press any key to continue...")
                return Result.Ok(True)
        if self.job_status == JobStatus.ERROR:
            result = self.failed_job_options_prompt()
            if result.failure:
                return Result.Ok(self.exit_menu)
            user_choice = result.value
            if user_choice == "RETRY":
                job_runner = JobRunner(
                    db_job=self.db_job,
                    db_session=self.db_session,
                    config=self.config,
                    scraped_data=self.scraped_data,
                )
                result = job_runner.execute()
                if result.failure:
                    return result
                return Result.Ok(True)
        else:
            pause(message="\nPress any key to return to the previous menu...")
            return Result.Ok(self.exit_menu)

    def incomplete_job_options_prompt(self):
        choices = {
            f"{MENU_NUMBERS.get(1)}  Execute Job": "RUN",
            f"{MENU_NUMBERS.get(2)}  Cancel Job": "CANCEL",
            f"{EMOJI_DICT.get('BACK')} Return to Incomplete Jobs": None,
        }
        prompt = Bullet(
            "Current options:",
            choices=[choice for choice in choices.keys()],
            bullet="",
            shift=1,
            indent=2,
            margin=2,
            bullet_color=colors.foreground["default"],
            background_color=colors.foreground["default"],
            background_on_switch=colors.foreground["default"],
            word_color=colors.foreground["default"],
            word_on_switch=colors.bright(colors.foreground["cyan"]),
        )
        choice_text = prompt.launch()
        choice_value = choices.get(choice_text)
        return Result.Ok(choice_value) if choice_value else Result.Fail("")

    def failed_job_options_prompt(self):
        choices = {
            f"{MENU_NUMBERS.get(1)}  Retry Job": "RETRY",
            f"{EMOJI_DICT.get('BACK')} Return to Failed Jobs": None,
        }
        prompt = Bullet(
            "Current options:",
            choices=[choice for choice in choices.keys()],
            bullet="",
            shift=1,
            indent=2,
            margin=2,
            bullet_color=colors.foreground["default"],
            background_color=colors.foreground["default"],
            background_on_switch=colors.foreground["default"],
            word_color=colors.foreground["default"],
            word_on_switch=colors.bright(colors.foreground["cyan"]),
        )
        choice_text = prompt.launch()
        choice_value = choices.get(choice_text)
        return Result.Ok(choice_value) if choice_value else Result.Fail("")
